#!/usr/bin/env python3

# subtask_scheduling.py
#
# Copyright (C) 2015
# ASTRON (Netherlands Institute for Radio Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.
#
# $Id: subtask_scheduling.py 1580 2015-09-30 14:18:57Z loose $

"""
The subtask_scheduling service schedules TMSS subtasks.
It listens on the lofar notification message bus for state changes of TMSS subtasks; when a task finished,
it schedules (rest action) all successors that are in state 'defined'.
"""

import os
import logging
import threading
from lofar.common.util import waitForInterrupt
from lofar.common.datetimeutils import round_to_second_precision
from datetime import datetime, timedelta
from dateutil import parser

logger = logging.getLogger(__name__)

# warning: we import the old qpid messagebus here. We have to, because cobalt/otdb still use it.
# TODO: introduce AMQP/RabbitMQ/Kombu messagebus in cobalt, get rid of otdb, then use the new messagebus here.
from lofar.messagebus import messagebus as old_qpid_messagebus

# TMSS already uses the new AMQP/RabbitMQ/Kombu messagebus. Use it here to listen for subtask status changes.
from lofar.sas.tmss.client.tmssbuslistener import TMSSBusListener, TMSSEventMessageHandler, DEFAULT_BROKER, DEFAULT_BUSNAME
from lofar.sas.tmss.client.tmss_http_rest_client import TMSSsession

class HybridFeedbackMessageHandler(TMSSEventMessageHandler):
    '''This new-style message handler listens to the TMSS event messages using the TMSSBusListener
    and reads/processes old-style qpid feedback messages.

    We need such a hybrid solution because TMSS gives us the subtask status changes,
    and qpid/cobalt/pipelines give us the feedback. This old-style qpid handling needs to be replace once cobalt/pipelines are adapted.'''

    # name of the qpid queue where the qpid feedback for dataprodcuts is received.
    QPID_DATAPRODUCT_FEEDBACK_QUEUE = "otdb.task.feedback.dataproducts.for_tmss"

    # wait this long before cancelling the subtask if not all feedback is received
    DEFAULT_FEEDBACK_WAIT_TIMEOUT = 3600

    def __init__(self, rest_client_creds_id: str=None, qpid_broker: str=old_qpid_messagebus.broker_feedback, feedback_wait_timeout: int=DEFAULT_FEEDBACK_WAIT_TIMEOUT) -> None:
        super().__init__(log_event_messages=False)
        # a dict of subtask_id -> wait_timeout_timestamp for timeout computation
        self._finishing_subtasks = {}
        self._tmss_client = TMSSsession.create_from_dbcreds_for_ldap(rest_client_creds_id)
        self._qpid_broker = qpid_broker
        self._old_qpid_frombus = None
        self._feedback_wait_timeout = feedback_wait_timeout

    def start_handling(self):
        self._tmss_client.open()
        try:
            self._old_qpid_frombus = old_qpid_messagebus.FromBus(self.QPID_DATAPRODUCT_FEEDBACK_QUEUE, broker=self._qpid_broker)
        except Exception as e:
            logger.warning("Could not connect to old-style qpid messagebus: %s", e)
        self._init_timeouts_for_finishing_subtasks()
        super().start_handling()

    def stop_handling(self):
        super().stop_handling()
        if self._old_qpid_frombus:
            self._old_qpid_frombus.close()
        self._tmss_client.close()

    def _init_wait_timeout_for_finishing_observation_or_pipeline_subtask(self, subtask: dict):
        if subtask['state_value'] == 'finishing':
            specifications_template = self._tmss_client.get_url_as_json_object(subtask['specifications_template'])
            if specifications_template['type_value'].lower() in ('observation', 'pipeline'):
                try:
                    finishing_timestamp = parser.parse(subtask['stop_time'], ignoretz=True)
                except:
                    finishing_timestamp = datetime.utcnow()
                wait_timeout_timestamp = round_to_second_precision(finishing_timestamp + timedelta(seconds=self._feedback_wait_timeout))
                logger.info('waiting at most %d seconds until %s before cancelling %s subtask id=%s if not all feedback is received by then',
                            (wait_timeout_timestamp - datetime.utcnow()).total_seconds(), wait_timeout_timestamp,
                            specifications_template['type_value'], subtask['id'])
                self._finishing_subtasks[subtask['id']] = wait_timeout_timestamp

    def _init_timeouts_for_finishing_subtasks(self):
        '''upon startup, initialize the timeout for all currently finishing subtasks. Allows for service restarts.'''
        subtasks = self._tmss_client.get_subtasks(state='finishing')
        for subtask in subtasks:
            self._init_wait_timeout_for_finishing_observation_or_pipeline_subtask(subtask)

    def onSubTaskStatusChanged(self, id: int, status: str):
        '''Handle TMSS subtask status changes'''
        logger.info("subtask id=%s status changed to %s", id, status)
        # keep track of finishing_status_change_timestamp for timeout computation
        if status == 'finishing':
            subtask = self._tmss_client.get_subtask(id)
            self._init_wait_timeout_for_finishing_observation_or_pipeline_subtask(subtask)
        elif status in ('finished', 'cancelling', 'cancelled', 'error'):
            if id in self._finishing_subtasks:
                wait_timeout_timestamp = self._finishing_subtasks[id]
                logger.info("removing remaining feedback wait time of %s seconds for subtask id=%s because the status is %s, ",
                            (wait_timeout_timestamp - datetime.utcnow()).total_seconds(), id, status)
                del self._finishing_subtasks[id]

    def before_receive_message(self):
        # use TMSSEventMessageHandler template pattern to act perform extra business logic in the loop
        self._read_feedback_message_and_process(1)
        self._set_subtask_status_to_cancelled_upon_feedback_timeout()
        super().before_receive_message()

    def _set_subtask_status_to_cancelled_upon_feedback_timeout(self):
        for subtask_id, wait_timeout_timestamp in list(self._finishing_subtasks.items()):
            if datetime.utcnow() > wait_timeout_timestamp:
                del self._finishing_subtasks[subtask_id]
                logger.warning('cancelling subtask id=%s due to timeout of %s seconds while waiting for feedback', subtask_id, self._feedback_wait_timeout)
                self._tmss_client.set_subtask_status(subtask_id, 'cancelling')
                self._tmss_client.set_subtask_status(subtask_id, 'cancelled')

    def _read_feedback_message_and_process(self, timeout: float=1):
        try:
            if self._old_qpid_frombus is None:
                return

            # get message from messagebus
            msg = self._old_qpid_frombus.get(timeout)

            if msg is not None:
                content = msg.content()
                logger.info("received message from qpid queue='%s' %s", self.QPID_DATAPRODUCT_FEEDBACK_QUEUE, content)
                self._old_qpid_frombus.ack(msg)

                # note: cobalt/rtcp creates feedback and assumes that the observationID has its origin in OTDB.
                # hence, it stores this id in the 'sasid' property of the message.
                # We know that TMSS sets its subtask_id in the parset in the Observation.ObsID field,
                # so we can fetch the TMSS subtask_id from the msg's sasid.
                tmss_subtask_id = content.sasid
                feedback = content.payload

                logger.info("feedback for TMSS subtask id=%s feedback=%s", tmss_subtask_id, feedback)

                self.process_feedback_and_set_to_finished_if_complete(tmss_subtask_id, feedback)
        except TimeoutError:
            pass
        except Exception as e:
            logger.error(str(e))

    def process_feedback_and_set_to_finished_if_complete(self, subtask_id: int, feedback: str):
        logger.info('submitting feedback for subtask id=%s to TMSS', subtask_id)
        updated_subtask = self._tmss_client.process_feedback_and_set_to_finished_if_complete(subtask_id=subtask_id, feedback=feedback)
        logger.info('subtask id=%s with the processed feedback has state %s', subtask_id, updated_subtask['state_value'])

def create_service(exchange: str = DEFAULT_BUSNAME, broker: str = DEFAULT_BROKER, qpid_broker: str=old_qpid_messagebus.broker_feedback, feedback_wait_timeout: int=HybridFeedbackMessageHandler.DEFAULT_FEEDBACK_WAIT_TIMEOUT, rest_client_creds_id: str=None):
    return TMSSBusListener(handler_type=HybridFeedbackMessageHandler, handler_kwargs={ "rest_client_creds_id": rest_client_creds_id,
                                                                                       "qpid_broker": qpid_broker,
                                                                                       "feedback_wait_timeout": feedback_wait_timeout },
                           exchange=exchange, broker=broker)

def main():
    # make sure we run in UTC timezone
    import os
    os.environ['TZ'] = 'UTC'

    logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=logging.INFO)

    from optparse import OptionParser, OptionGroup

    # Check the invocation arguments
    parser = OptionParser('%prog [options]',
                          description='run the tmss_feedback_handling_service which relays observation feedback from the old QPID messaging system into TMSS.')

    parser.add_option('-t', '--timeout', dest='timeout', type='int', default=HybridFeedbackMessageHandler.DEFAULT_FEEDBACK_WAIT_TIMEOUT,
                      help='Wait for <timeout> seconds before a subtask is cancelled if not all feedback is received, default: %default')

    group = OptionGroup(parser, 'QPID Messaging options',
                        description="This feedback service connects to the (old and almost obsolete) QPID broker. This can and should be replaced by connecting to RabbitMQ when Cobalt and MAC have been adapted.")
    group.add_option('-q', '--qpid_broker', dest='qpid_broker', type='string', default=old_qpid_messagebus.broker_feedback, help='Address of the QPID broker, default: %default')
    parser.add_option_group(group)

    group = OptionGroup(parser, 'RabbitMQ Messaging options')
    group.add_option('-b', '--broker', dest='broker', type='string', default=DEFAULT_BROKER, help='Address of the message broker, default: %default')
    group.add_option('-e', "--exchange", dest="exchange", type="string", default=DEFAULT_BUSNAME, help="exchange where the TMSS event messages are published. [default: %default]")
    parser.add_option_group(group)

    group = OptionGroup(parser, 'TMSS Django options')
    parser.add_option_group(group)
    group.add_option('-R', '--rest_credentials', dest='rest_credentials', type='string', default='TMSSClient', help='TMSS Django REST API credentials name, default: %default')

    (options, args) = parser.parse_args()

    TMSSsession.check_connection_and_exit_on_error(options.rest_credentials)

    with create_service(exchange=options.exchange, broker=options.broker, rest_client_creds_id=options.rest_credentials, qpid_broker=options.qpid_broker, feedback_wait_timeout=options.timeout):
        waitForInterrupt()

if __name__ == '__main__':
    main()
