#!/usr/bin/env python3

# Copyright (C) 2012-2015  ASTRON (Netherlands Institute for Radio Astronomy)
# P.O. Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.

import logging
logger = logging.getLogger(__name__)

import os
import threading
import datetime
from datetime import timedelta
import time
from lofar.common.util import waitForInterrupt

# Default values of parameters
INTERVAL_TIME_SECONDS = 24 * 60 * 60  # 24 hours (every day one calculation ahead)
NBR_DAYS_CALCULATE_AHEAD = 365    # 1 year
NBR_DAYS_BEFORE_TODAY = 1


def execute_populate_sunrise_and_sunset_for_all_stations(nbr_days_calculate_ahead, start_date):
    """
    Execute the populate of calculations (sunrise/sunset) for given number of days stating at give date
    :param nbr_days_calculate_ahead: Number of days to calculated
    :param start_date: The date to start calculate
    :return next_date: The next_date to process
    """
    logger.info("execute_populate_sunrise_and_sunset_for_all_stations %s for %d days" % (start_date, nbr_days_calculate_ahead))
    # Import here otherwise you get
    # "django.core.exceptions.ImproperlyConfigured: Requested setting INSTALLED_APPS, but settings are not configured. You must either define the environment variable DJANGO_SETTINGS_MODULE or call settings.configure() before accessing settings."
    from lofar.sas.tmss.tmss.tmssapp.populate import populate_sunrise_and_sunset_for_all_stations

    populate_sunrise_and_sunset_for_all_stations(nbr_days=nbr_days_calculate_ahead, start_date=start_date)
    # Return the next_date to process
    next_date = start_date + datetime.timedelta(days=nbr_days_calculate_ahead)
    return next_date


class TMSSPreCalculationsServiceJob(threading.Thread):
    def __init__(self, interval, execute, *args, **kwargs):
        threading.Thread.__init__(self)
        self.daemon = False
        self.stopped = threading.Event()
        self.interval = interval
        self.execute = execute
        self.args = args
        self.kwargs = kwargs

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def stop(self):
        self.stopped.set()
        self.join()

    def run(self):
        start_time = time.time()
        next_date = self.execute(*self.args, **self.kwargs)
        # determine remaining time for exact heartbeat of the interval time
        remaining_wait_time_in_sec = self.interval.total_seconds() - (time.time() - start_time)
        while not self.stopped.wait(remaining_wait_time_in_sec):
            self.kwargs["nbr_days_calculate_ahead"] = 1
            self.kwargs["start_date"] = next_date
            start_time = time.time()
            next_date = self.execute(*self.args, **self.kwargs)
            remaining_wait_time_in_sec = self.interval.total_seconds() - (time.time() - start_time)


def create_service_job_for_sunrise_and_sunset_calculations(interval_time, nbr_days_calculate_ahead, nbr_days_before_today):
    start_date = datetime.date.today() - datetime.timedelta(days=nbr_days_before_today)
    return TMSSPreCalculationsServiceJob(interval=timedelta(seconds=interval_time),
                                         execute=execute_populate_sunrise_and_sunset_for_all_stations,
                                         nbr_days_calculate_ahead=nbr_days_calculate_ahead, start_date=start_date)


def main():
    # make sure we run in UTC timezone
    os.environ['TZ'] = 'UTC'

    from optparse import OptionParser, OptionGroup
    from lofar.common import dbcredentials

    logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=logging.INFO)

    # Check the invocation arguments
    parser = OptionParser('%prog [options]',
                          description='run the tmss_workflow_service which forwards TMSS events to the workflow engine.')

    parser.add_option('-i', '--interval_time', dest='interval_time', type='int', default=INTERVAL_TIME_SECONDS,
                      help='The time between next calculation, default: %default')
    parser.add_option('-d', '--nbr_days_calculate_ahead', dest='nbr_days_calculate_ahead', type='int', default=NBR_DAYS_CALCULATE_AHEAD,
                      help='The number of days to calculate the sunset/sunrise ahead, default: %default')
    parser.add_option('-b', '--nbr_days_before_today', dest='nbr_days_before_today', type='int', default=NBR_DAYS_BEFORE_TODAY,
                      help='The number of days to calculate the sunset/sunrise before today (so yesterday=1), default: %default')

    group = OptionGroup(parser, 'Django options')
    parser.add_option_group(group)
    group.add_option('-C', '--credentials', dest='dbcredentials', type='string', default=os.environ.get('TMSS_DBCREDENTIALS', 'TMSS'), help='django dbcredentials name, default: %default')

    (options, args) = parser.parse_args()
    from lofar.sas.tmss.tmss import setup_and_check_tmss_django_database_connection_and_exit_on_error
    setup_and_check_tmss_django_database_connection_and_exit_on_error(options.dbcredentials)

    job = create_service_job_for_sunrise_and_sunset_calculations(options.interval_time, options.nbr_days_calculate_ahead, options.nbr_days_before_today)
    job.start()
    waitForInterrupt()
    job.stop()


if __name__ == '__main__':
    main()

