#!/usr/bin/env python3

# Copyright (C) 2020
# ASTRON (Netherlands Institute for Radio Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.
#

"""
This __init__ module for this constraints python package defines the 'API' to:
 - filter a list of schedulable scheduling_units by checking their constraints: see method filter_scheduling_units_using_constraints
 - sort a (possibly filtered) list of schedulable scheduling_units evaluating their constraints and computing a 'fitness' score: see method get_sorted_scheduling_units_scored_by_constraints
These main methods are used in the dynamic_scheduler to pick the next best scheduling unit, and compute the midterm schedule.

Currently we have only one SchedulingConstraintsTemplate in TMSS, named 'constraints', version 1.
But, it is envisioned that we we get more templates.
So, based on the template the actual filter and score methods are selected from a specific module.
By convention we use one module per template. Currently, we have and use only module template_constraints_v1.py

If/When we add a new SchedulingConstraintsTemplate, then we should add a new module with the specific filter and score methods,
and add a extra 'if' in the strategy pattern used here. (see below for implementation)
"""

import logging
logger = logging.getLogger(__name__)
from datetime import datetime
from typing import NamedTuple

from lofar.sas.tmss.tmss.tmssapp import models
from lofar.sas.tmss.tmss.exceptions import *
from lofar.sas.tmss.tmss.tmssapp.reservations import get_active_station_reservations_in_timewindow

################## main data struct and methods ##################

class ScoredSchedulingUnit(NamedTuple):
    '''struct for collecting scores per constraint and a weighted_score for a scheduling_unit at the given start_time
    '''
    scheduling_unit: models.SchedulingUnitBlueprint
    scores: dict
    start_time: datetime
    weighted_score: float


def filter_scheduling_units_using_constraints(scheduling_units: [models.SchedulingUnitBlueprint], lower_bound: datetime, upper_bound: datetime) -> [models.SchedulingUnitBlueprint]:
    """
    Filter the given scheduling_units by whether their constraints are met within the given timewindow.
    If one or more scheduling units can run only within this time window and not after it, then only these exclusivly runnable scheduling units.
    :param lower_bound: evaluate and score the constrains at and after lower_bound_start_time. The returned unit has a start_time guaranteed at or after lower_bound_start_time.
    :param upper_bound: evaluate and score the constrains before upper_bound_stop_time. The returned unit has a stop_time guaranteed before upper_bound_stop_time.
    :param scheduling_units: evaluate/filter these scheduling_units.
    Returns a list scheduling_units for which their constraints are met within the given timewindow.
    """
    runnable_scheduling_units = []
    runnable_exclusive_in_this_window_scheduling_units = []

    for scheduling_unit in scheduling_units:
        try:
            if scheduling_unit.draft is None or scheduling_unit.draft.scheduling_constraints_template is None:
                logger.warning("cannot dynamically schedule scheduling_unit id=%s name='%s' because it has not constraints template", scheduling_unit.id, scheduling_unit.name)
                continue

            if can_run_within_timewindow(scheduling_unit, lower_bound, upper_bound) and can_run_within_station_reservations(scheduling_unit):
                runnable_scheduling_units.append(scheduling_unit)

                # if a schedulingunit cannot run after this window, then apparently its limited to run exclusively in this time window.
                earliest_possible_start_time = get_earliest_possible_start_time(scheduling_unit, lower_bound)
                if not can_run_after(scheduling_unit, earliest_possible_start_time+scheduling_unit.duration):
                    runnable_exclusive_in_this_window_scheduling_units.append(scheduling_unit)
        except UnknownTemplateException as e:
            # TODO: how do we notify the user that we cannot dynamically schedule this sub due to an unknown template?
            # current pragmatic solution: log warning, and set sub state to error via its schedulable subtasks.
            # This ensures that the unit is not schedulable anymore, and forces the user to take action.
            # For example, the user can choose a different template,
            # or submit a feature request to implement constraint solvers for this new template.
            logger.warning(e)
            for subtask in models.Subtask.independent_subtasks().filter(task_blueprint__scheduling_unit_blueprint_id=scheduling_unit.id).all():
                subtask.status = models.SubtaskState.objects.get(value=models.SubtaskState.Choices.ERROR.value)
                subtask.save()

    # if we have schedulingunit(s) that can run exclusively in this time window (and not afterwards), then return only these.
    if runnable_exclusive_in_this_window_scheduling_units:
        return runnable_exclusive_in_this_window_scheduling_units

    # there are no exclusive units, so return all runnable_scheduling_units
    return runnable_scheduling_units


def get_best_scored_scheduling_unit_scored_by_constraints(scheduling_units: [models.SchedulingUnitBlueprint], lower_bound_start_time:datetime, upper_bound_stop_time:datetime) -> ScoredSchedulingUnit:
    """
    get the best scored schedulable scheduling_unit which can run withing the given time window from the given scheduling_units.
    :param lower_bound_start_time: evaluate and score the constrains at and after lower_bound_start_time. The returned unit has a start_time guaranteed at or after lower_bound_start_time.
    :param upper_bound_stop_time: evaluate and score the constrains before upper_bound_stop_time. The returned unit has a stop_time guaranteed before upper_bound_stop_time.
    :param scheduling_units: evaluate these scheduling_units.
    Returns a ScoredSchedulingUnit struct with the best next schedulable scheduling unit and its proposed start_time where it best fits its contraints.
    """
    sorted_scored_scheduling_units = sort_scheduling_units_scored_by_constraints(scheduling_units, lower_bound_start_time, upper_bound_stop_time)

    if sorted_scored_scheduling_units:
        # they are sorted best to worst, so return/use first.
        best_scored_scheduling_unit = sorted_scored_scheduling_units[0]
        return best_scored_scheduling_unit

    return None


def sort_scheduling_units_scored_by_constraints(scheduling_units: [models.SchedulingUnitBlueprint], lower_bound_start_time: datetime, upper_bound_stop_time: datetime) -> [ScoredSchedulingUnit]:
    """
    Compute the score and proposed start_time for all given scheduling_units. Return them sorted by their weighted_score.
    :param lower_bound_start_time: evaluate and score the constrains at and after lower_bound_start_time. The returned unit has a start_time guaranteed at or after lower_bound_start_time.
    :param upper_bound_stop_time: evaluate and score the constrains before upper_bound_stop_time. The returned unit has a stop_time guaranteed before upper_bound_stop_time.
    :param scheduling_units: evaluate these scheduling_units.
    Returns a list of ScoredSchedulingUnit structs with the score details, a weighted_score and a proposed start_time where it best fits its contraints.
    """

    scored_scheduling_units = []
    for scheduling_unit in scheduling_units:
        try:
            scored_scheduling_unit = compute_scores(scheduling_unit, lower_bound_start_time, upper_bound_stop_time)

            # check and ensure that the proposed start_time is within the required [lower_bound_start_time, upper_bound_stop_time] window.
            schedulable_unit = scored_scheduling_unit.scheduling_unit
            proposed_start_time = scored_scheduling_unit.start_time
            proposed_stop_time = proposed_start_time + schedulable_unit.duration

            if proposed_start_time < lower_bound_start_time:
                raise DynamicSchedulingException("The best next schedulable scheduling_unit id=%s has a proposed start_time '%s' before the given lower bound '%s'" % (
                    schedulable_unit.id, proposed_start_time, lower_bound_start_time))

            if proposed_stop_time > upper_bound_stop_time:
                raise DynamicSchedulingException("The best next schedulable scheduling_unit id=%s has a proposed stop_time '%s' after the given upper bound '%s'" % (
                    schedulable_unit.id, proposed_stop_time, upper_bound_stop_time))

            scored_scheduling_units.append(scored_scheduling_unit)
        except (UnknownTemplateException, DynamicSchedulingException) as e:
            # TODO: how do we notify the user that we cannot dynamically schedule this sub due to an unknown template?
            # current pragmatic solution: log warning, and set sub state to error via its schedulable subtasks.
            # This ensures that the unit is not schedulable anymore, and forces the user to take action.
            # For example, the user can choose a different template,
            # or submit a feature request to implement constraint solvers for this new template.
            logger.warning(e)
            for subtask in models.Subtask.independent_subtasks().filter(task_blueprint__scheduling_unit_blueprint_id=scheduling_unit.id).all():
                subtask.status = models.SubtaskState.objects.get(value=models.SubtaskState.Choices.ERROR.value)
                subtask.save()

    return sorted(scored_scheduling_units, key=lambda x: x.weighted_score, reverse=True)


################## helper methods #################################################################
#                                                                                                 #
# these helper methods are selected by a strategy pattern based on the template name and version  #
# The actual implementation can be found in the other module(s) in this package                   #
# Currently we only have one template with one implementation in template_constraints_v1.py       #
#                                                                                                 #
###################################################################################################

def can_run_within_timewindow(scheduling_unit: models.SchedulingUnitBlueprint, lower_bound: datetime, upper_bound: datetime) -> bool:
    '''Check if the given scheduling_unit can run somewhere within the given time window depending on the sub's constrains-template/doc.'''
    constraints_template = scheduling_unit.draft.scheduling_constraints_template

    # choose appropriate method based on template (strategy pattern), or raise
    if constraints_template.name == 'constraints' and constraints_template.version == 1:
        # import here to prevent circular imports. Do not worry about performance loss, cause python only imports once and then uses a cache.
        from . import template_constraints_v1
        return template_constraints_v1.can_run_within_timewindow(scheduling_unit, lower_bound, upper_bound)

    # TODO: if we get more constraint templates or versions, then add a check here and import and use the new module with the constraint methods for that specific template. (strategy pattern)

    raise UnknownTemplateException("Cannot check if scheduling_unit id=%s can run between '%s' and '%s', because we have no constraint checker for scheduling constraints template '%s' version=%s" % (
                                    scheduling_unit.id, lower_bound, upper_bound, constraints_template.name, constraints_template.version))


def can_run_after(scheduling_unit: models.SchedulingUnitBlueprint, lower_bound: datetime) -> bool:
    '''Check if the given scheduling_unit can run somewhere after the given lowerbound timestamp depending on the sub's constrains-template/doc.'''
    constraints_template = scheduling_unit.draft.scheduling_constraints_template

    # choose appropriate method based on template (strategy pattern), or raise
    if constraints_template.name == 'constraints' and constraints_template.version == 1:
        # import here to prevent circular imports. Do not worry about performance loss, cause python only imports once and then uses a cache.
        from . import template_constraints_v1
        return template_constraints_v1.can_run_after(scheduling_unit, lower_bound)

    # TODO: if we get more constraint templates or versions, then add a check here and import and use the new module with the constraint methods for that specific template. (strategy pattern)

    raise UnknownTemplateException("Cannot check if scheduling_unit id=%s can run after '%s', because we have no constraint checker for scheduling constraints template '%s' version=%s" % (
                                    scheduling_unit.id, lower_bound, constraints_template.name, constraints_template.version))



def compute_scores(scheduling_unit: models.SchedulingUnitBlueprint, lower_bound:datetime, upper_bound:datetime) -> ScoredSchedulingUnit:
    '''Compute the "fitness" scores per constraint for the given scheduling_unit at the given starttime depending on the sub's constrains-template/doc.'''
    constraints_template = scheduling_unit.draft.scheduling_constraints_template

    # choose appropriate method based on template (strategy pattern), or raise
    if constraints_template.name == 'constraints' and constraints_template.version == 1:
        # import here to prevent circular imports. Do not worry about performance loss, cause python only imports once and then uses a cache.
        from . import template_constraints_v1
        return template_constraints_v1.compute_scores(scheduling_unit, lower_bound, upper_bound)

    # TODO: if we get more constraint templates or versions, then add a check here and import and use the new module with the constraint methods for that specific template. (strategy pattern)

    raise UnknownTemplateException("Cannot compute scores for scheduling_unit id=%s, because we have no score computation method for scheduling constraints template '%s' version=%s" % (
                                    scheduling_unit.id, constraints_template.name, constraints_template.version))


def get_earliest_possible_start_time(scheduling_unit: models.SchedulingUnitBlueprint, lower_bound: datetime) -> datetime:
    '''determine the earliest possible start_time for the given scheduling unit, taking into account all its constraints'''
    constraints_template = scheduling_unit.draft.scheduling_constraints_template

    # choose appropriate method based on template (strategy pattern), or raise
    if constraints_template.name == 'constraints' and constraints_template.version == 1:
        # import here to prevent circular imports. Do not worry about performance loss, cause python only imports once and then uses a cache.
        from . import template_constraints_v1
        return template_constraints_v1.get_earliest_possible_start_time(scheduling_unit, lower_bound)

    # TODO: if we get more constraint templates or versions, then add a check here and import and use the new module with the constraint methods for that specific template. (strategy pattern)

    raise UnknownTemplateException("Cannot compute earliest possible start_time for scheduling_unit id=%s, because we have no constraint checker for scheduling constraints template '%s' version=%s" % (
                                    scheduling_unit.id, constraints_template.name, constraints_template.version))


def get_min_earliest_possible_start_time(scheduling_units: [models.SchedulingUnitBlueprint], lower_bound: datetime) -> datetime:
    '''deterimine the earliest possible starttime over all given scheduling units, taking into account all their constraints'''
    try:
        return min(get_earliest_possible_start_time(scheduling_unit, lower_bound) for scheduling_unit in scheduling_units if scheduling_unit.draft.scheduling_constraints_template is not None)
    except ValueError:
        return lower_bound





def can_run_within_station_reservations(scheduling_unit: models.SchedulingUnitBlueprint) -> bool:
    """
    Check if the given scheduling_unit can run if the reserved station are taken into account.
    The station requirement will be evaluated. If a reserved station will be used within the time window of
    the given boundaries (start/stop time) for this scheduling unit then this function will return False.
    """
    can_run = True
    # Get a station list of given SchedulingUnitBlueprint
    lst_stations_to_be_used = scheduling_unit.flat_station_list

    sub_start_time = scheduling_unit.start_time
    sub_stop_time = scheduling_unit.stop_time

    lst_reserved_stations = get_active_station_reservations_in_timewindow(sub_start_time, sub_stop_time)
    # Check if the reserved stations are going to be used
    common_set_stations = set(lst_stations_to_be_used).intersection(lst_reserved_stations)
    if len(common_set_stations) > 0:
        logger.warning("There is/are station(s) reserved %s which overlap with timewindow  [%s - %s]",
                       common_set_stations, sub_start_time, sub_stop_time)
        # Check which stations are in overlap/common per station group. If more than max_nr_missing stations
        # are in overlap then can_run is actually false, otherwise it is still within policy and ok
        station_groups = scheduling_unit.station_groups
        for sg in station_groups:
            nbr_missing = len(set(sg["stations"]) & set(common_set_stations))
            if "max_nr_missing" in sg:
                max_nr_missing = sg["max_nr_missing"]
            else:
                max_nr_missing = 0
            if nbr_missing > max_nr_missing:
                logger.info("There are more stations in reservation than the specification is given "
                            "(%d is larger than %d). The stations that are in conflict are '%s'."
                            "Can not run scheduling_unit id=%d " %
                               (nbr_missing, max_nr_missing, common_set_stations, scheduling_unit.pk))
                can_run = False
                break
    return can_run

