#!/usr/bin/env python3

# subtask_scheduling.py
#
# Copyright (C) 2015
# ASTRON (Netherlands Institute for Radio Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.
#
# $Id: subtask_scheduling.py 1580 2015-09-30 14:18:57Z loose $

"""
The subtask_scheduling service schedules TMSS subtasks.
It listens on the lofar notification message bus for state changes of TMSS subtasks; when a task finished,
it schedules (rest action) all successors that are in state 'defined'.
"""

import os
from optparse import OptionParser
import logging
logger = logging.getLogger(__name__)

from lofar.sas.tmss.client.tmssbuslistener import *
from lofar.sas.tmss.client.tmss_http_rest_client import TMSSsession

class TMSSSubTaskSchedulingEventMessageHandler(TMSSEventMessageHandler):
    '''
    '''
    def __init__(self, tmss_client_credentials_id: str=None):
        super().__init__()
        self.tmss_client = TMSSsession.create_from_dbcreds_for_ldap(tmss_client_credentials_id)

    def start_handling(self):
        self.tmss_client.open()
        super().start_handling()

    def stop_handling(self):
        super().stop_handling()
        self.tmss_client.close()

    def onSubTaskStatusChanged(self, id: int, status: str):
        super().onSubTaskStatusChanged(id, status)

        if status == "finished":
            successors = self.tmss_client.get_subtask_successors(id, state="defined")
            successor_ids = sorted([s['id'] for s in successors])

            logger.info("subtask %s finished. trying to schedule defined successors: %s",
                        id, ', '.join(str(id) for id in successor_ids) or 'None')

            for successor in successors:
                try:
                    suc_subtask_id = successor['id']
                    suc_subtask_state = successor['state_value']

                    if suc_subtask_state == "defined":
                        successor_predecessors = self.tmss_client.get_subtask_predecessors(suc_subtask_id)

                        if any([suc_pred['state_value']!='finished' for suc_pred in successor_predecessors]):
                            logger.info("skipping scheduling of successor subtask %s for finished subtask %s because not all its other predecessor subtasks are finished", suc_subtask_id, id)
                        else:
                            logger.info("trying to schedule successor subtask %s for finished subtask %s", suc_subtask_id, id)
                            # try scheduling the subtask.
                            # if it succeeds, then the state will be 'scheduled' afterwards
                            # if there is a specification error, then the state will be 'error' afterwards
                            # if there is another kind of error (like needing ingest-permission), then the state will be 'defined' afterwards, so you can retry.
                            #   for the ingest-permission we will retry automatically when that permission is granted
                            scheduled_successor = self.tmss_client.schedule_subtask(suc_subtask_id)
                            suc_subtask_state = scheduled_successor['state_value']
                            logger.log(logging.INFO if suc_subtask_state=='scheduled' else logging.WARNING,
                                       "successor subtask %s for finished subtask %s now has state '%s', see %s", suc_subtask_id, id, suc_subtask_state, scheduled_successor['url'])
                    else:
                        logger.warning("skipping scheduling of successor subtask %s for finished subtask %s because its state is '%s'", suc_subtask_id, id, suc_subtask_state)

                except Exception as e:
                    logger.error(e)

    def onSchedulingUnitBlueprintIngestPermissionGranted(self, id: int, ingest_permission_granted_since: datetime):
        logger.info("ingest_permission_granted_since='%s' for scheduling_unit_blueprint id=%s", ingest_permission_granted_since, id)
        scheduling_unit = self.tmss_client.get_path_as_json_object("scheduling_unit_blueprint_extended/%s" % id)
        for task in scheduling_unit['task_blueprints']:
            for subtask in task['subtasks']:
                if subtask['state_value'] == 'defined':
                    subtask_template = self.tmss_client.get_url_as_json_object(subtask['specifications_template'])
                    if subtask_template['type_value'] == 'ingest':
                        logger.info("trying to schedule ingest subtask id=%s for scheduling_unit_blueprint id=%s...", subtask['id'], id)
                        self.tmss_client.schedule_subtask(subtask['id'])


def create_subtask_scheduling_service(exchange: str=DEFAULT_BUSNAME, broker: str=DEFAULT_BROKER, tmss_client_credentials_id: str=None):
    return TMSSBusListener(handler_type=TMSSSubTaskSchedulingEventMessageHandler,
                                  handler_kwargs={'tmss_client_credentials_id': tmss_client_credentials_id},
                                  exchange=exchange,
                                  broker=broker)

def main():
    # make sure we run in UTC timezone
    os.environ['TZ'] = 'UTC'

    logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=logging.INFO)

    # Check the invocation arguments
    parser = OptionParser('%prog [options]',
                          description='run the tmss_subtask_scheduling_service which automatically schedules the defined successor tasks for finished subtasks')
    parser.add_option('-q', '--broker', dest='broker', type='string', default=DEFAULT_BROKER, help='Address of the messaging broker, default: %default')
    parser.add_option('--exchange', dest='exchange', type='string', default=DEFAULT_BUSNAME, help='Name of the exchange on the messaging broker, default: %default')
    parser.add_option('-t', '--tmss_client_credentials_id', dest='tmss_client_credentials_id', type='string',
                      default=os.environ.get("TMSS_CLIENT_DBCREDENTIALS", "TMSSClient"),
                      help='the credentials id for the file in ~/.lofar/dbcredentials which holds the TMSS http REST api url and credentials, default: %default')
    (options, args) = parser.parse_args()

    with create_subtask_scheduling_service(options.exchange, options.broker, options.tmss_client_credentials_id):
        waitForInterrupt()

if __name__ == '__main__':
    main()
