from lofar.sas.tmss.tmss.exceptions import *
from lofar.sas.tmss.tmss.tmssapp.models.scheduling import Dataproduct, SubtaskType, Subtask, SubtaskOutput, SIPidentifier, Algorithm
from lofar.sas.tmss.tmss.tmssapp.models.specification import Datatype, Dataformat
from lofar.lta.sip import siplib, ltasip, validator, constants
from lofar.common.json_utils import add_defaults_to_json_object_for_schema

import uuid
import logging
import isodate
import datetime
logger = logging.getLogger(__name__)

mapping_antennaset_type_TMSS_2_SIP = {
    "HBA_DUAL": constants.ANTENNASETTYPE_HBA_DUAL,
    "HBA_DUAL_INNER": constants.ANTENNASETTYPE_HBA_DUAL_INNER,
    "HBA_JOINED": constants.ANTENNASETTYPE_HBA_JOINED,
    "HBA_JOINED_INNER": constants.ANTENNASETTYPE_HBA_JOINED_INNER,
    "HBA_ONE": constants.ANTENNASETTYPE_HBA_ONE,
    "HBA_ONE_INNER": constants.ANTENNASETTYPE_HBA_ONE_INNER,
    "HBA_ZERO": constants.ANTENNASETTYPE_HBA_ZERO,
    "HBA_ZERO_INNER": constants.ANTENNASETTYPE_HBA_ZERO_INNER,
    "LBA_INNER": constants.ANTENNASETTYPE_LBA_INNER,
    "LBA_OUTER": constants.ANTENNASETTYPE_LBA_OUTER,
    "LBA_SPARSE_EVEN": constants.ANTENNASETTYPE_LBA_SPARSE_EVEN,
    "LBA_SPARSE_ODD": constants.ANTENNASETTYPE_LBA_SPARSE_ODD,
    "LBA_ALL": "NOT SUPPORTED IN LTA YET"
}

mapping_filterset_type_TMSS_2_SIP = {
    "LBA_10_90": constants.FILTERSELECTIONTYPE_10_90_MHZ,
    "LBA_30_90": constants.FILTERSELECTIONTYPE_30_90_MHZ,
    "HBA_110_190": constants.FILTERSELECTIONTYPE_110_190_MHZ,
    "HBA_210_250": constants.FILTERSELECTIONTYPE_210_250_MHZ
}


def get_number_of_dataproducts_of_type(subtask, dataproduct_datatype):
    """
    Retrieve the number of dataproducts of given data type in subtask:
    :param subtask:
    :param dataproduct_datatype: Datatype like "visibilities", "time series", "instrument model", "image", "quality"
    :return:
    """
    nbr_dataproduct = 0
    subtask_outputs = list(SubtaskOutput.objects.filter(subtask_id=subtask.id))
    for subtask_output in subtask_outputs:
        dataproducts = list(Dataproduct.objects.filter(producer_id=subtask_output.id,
                                                       datatype=dataproduct_datatype))
        nbr_dataproduct += len(dataproducts)
    return nbr_dataproduct


def get_siplib_stations_list(subtask):
    """
    Retrieve a list of siplib Stations objects which is extracted from the station_list and the antennaset
    :param subtask:
    :return:

    Conversion logic see here: https://support.astron.nl/confluence/display/TMSS/Dataproduct+Provenance
    """
    siplib_station_list = []
    list_stations = subtask.specifications_doc['stations']['station_list']
    antennaset = subtask.specifications_doc['stations']['antenna_set']
    for station in list_stations:
        if antennaset.startswith('LBA'):
            antennafieldtypes = ['LBA']
        elif antennaset.startswith('HBA') and not station.startswith('CS'):
            antennafieldtypes = ['HBA']
        elif antennaset.startswith('HBA_DUAL'):
            antennafieldtypes = ['HBA0', 'HBA1']
        elif antennaset.startswith('HBA_ZERO'):
            antennafieldtypes = ['HBA0']
        elif antennaset.startswith('HBA_ONE'):
            antennafieldtypes = ['HBA1']
        else:
            raise ValueError('Cannot determine antennafields for station=%s antennaset=%s' % (station, antennaset))
        siplib_station_list.append(siplib.Station.preconfigured(station, antennafieldtypes))
    return siplib_station_list


def get_siplib_identifier(sipid_obj: SIPidentifier, context_str="") -> siplib.Identifier:
    """
    Retrieve an Identifier object. Get the unique_identifier and source of the given sip object and covert that to
    a siblib object
    :param sipid_obj: SIP Identifier object
    :param context_str: In case of failure add some context info
    :return: A lofar.lta.sip.siplib.Identifier object
    """
    identifier = siplib.Identifier.__new__(siplib.Identifier)
    unique_id = sipid_obj.unique_identifier
    if unique_id is None:
        raise TMSSException("Can not create SIP Identifier, the value of 'unique_id' is None for %s" % context_str)
    source = sipid_obj.source
    if source is None:
        raise TMSSException("Can not create SIP Identifier, the value 'source' is None for %s" % context_str)

    identifier._set_pyxb_identifier(
        ltasip.IdentifierType(
            source=source,
            identifier=unique_id,
            name=None,
            label=None),
        suppress_warning=True)
    return identifier


# todo: TMSS defines relations on task level. Technically, I guess it is more correct to go to subtask level when
#  adding provenance info with all intermediate dataproducts. So in case a task consists of several chaines processes,
#  you get the individual applies steps described in the SIP. Clarify that this is really what we want.
def add_subbtask_and_input_dataproducts_to_sip(subtask: Subtask, sip: siplib.Sip):
    """
    add a subtask including its input dataproducts to the sip, according to the specific observation or pipeline type they represent
    :param subtask: the Subtask object to add to the sip
    :param sip: the Sip object to add to
    """
    logger.info("Adding subtask id %s of type %s to sip." % (subtask.id, subtask.specifications_template.type.value))
    sip_subtask = create_sip_representation_for_subtask(subtask)
    if isinstance(sip_subtask, siplib.Observation):
        sip.add_observation(observation=sip_subtask)
    else:
        sip.add_pipelinerun(pipeline=sip_subtask)
        for input in subtask.inputs.all():
            for dataproduct in input.dataproducts.all():
                sip_dataproduct = create_sip_representation_for_dataproduct(dataproduct)
                sip.add_related_dataproduct(sip_dataproduct, return_xml=False)


def create_sip_representation_for_subtask(subtask: Subtask):
    """
    Extract info from the TMSS subtask and create a sip object representation corresponding to the type of the subtask
    :param subtask:
    :return: A siplib.Observation object or one of the various siplib pipeline object flavors
    """
    # determine common properties
    subtask_sip_identifier = get_siplib_identifier(subtask.global_identifier, "Subtask id=%s" % subtask.id)
    name = str(subtask.id)
    process_map = siplib.ProcessMap(strategyname=subtask.specifications_template.name,
                                    strategydescription=subtask.specifications_template.description,
                                    starttime=subtask.start_time,
                                    duration=isodate.duration_isoformat(datetime.timedelta(seconds=round((subtask.stop_time-subtask.start_time).total_seconds()))),
                                    identifier=subtask_sip_identifier,
                                    observation_identifier=subtask_sip_identifier,
                                    relations=[])  # todo, not sure this is still needed, can be empty for now (potentially scheduling_unit?)

    # determine subtask specific properties and add subtask representation to Sip object
    if subtask.specifications_template.type.value == SubtaskType.Choices.OBSERVATION.value:
        subarraypointings=None  # todo, subtask.specifications_doc, probably more complex than it looks -> RGOE yes complex type for later -> JK: assuming this is done in TMSS-308?
        observation = siplib.Observation(observingmode=constants.OBSERVINGMODETYPE_BEAM_OBSERVATION,  # can be hardcoded for an observation
                                         instrumentfilter=mapping_filterset_type_TMSS_2_SIP[subtask.specifications_doc['stations']['filter']],
                                         clock_frequency="200",  # fixed,
                                         clock_frequencyunit=constants.FREQUENCYUNIT_MHZ,  # fixed, consider Hz?
                                         stationselection=constants.STATIONSELECTIONTYPE_CUSTOM,  # fixed
                                         antennaset=mapping_antennaset_type_TMSS_2_SIP[subtask.specifications_doc['stations']['antenna_set']],
                                         timesystem="UTC",  # fixed
                                         numberofstations=len(subtask.specifications_doc['stations']['station_list']),
                                         stations=get_siplib_stations_list(subtask),
                                         numberofsubarraypointings=len(subtask.specifications_doc['stations']['digital_pointings']),
                                         numberoftbbevents=0,  # fixed
                                         numberofcorrelateddataproducts=get_number_of_dataproducts_of_type(subtask, Dataformat.Choices.MEASUREMENTSET.value),
                                         numberofbeamformeddataproducts=get_number_of_dataproducts_of_type(subtask, Dataformat.Choices.BEAMFORMED.value),
                                         numberofbitspersample=8,  # fixed
                                         process_map=process_map,
                                         channelwidth_frequency=None, # NA any more ('BlueGene compatibility' see comment in LTA-SIP.xsd)
                                         channelwidth_frequencyunit=constants.FREQUENCYUNIT_HZ,  # fixed
                                         observationdescription=subtask.task_blueprint.description,
                                         channelspersubband=0,  # NA any more ('BlueGene compatibility' see comment in LTA-SIP.xsd)
                                         subarraypointings=subarraypointings,
                                         transientbufferboardevents=None  # fixed
        )

        return observation
    elif subtask.specifications_template.type.value == SubtaskType.Choices.PIPELINE.value:
        sourcedata_identifiers = []
        for input in subtask.inputs.all():
            sourcedata_identifiers += [get_siplib_identifier(dp.global_identifier, "Dataproduct id=%s" % dp.id) for dp in input.dataproducts.all()]     # todo: use correct id, lookup based on TMSS reference or so, tbd
        if not sourcedata_identifiers:
            raise TMSSException("There seems to be no subtask input associated to your pipeline subtask id %s. Please define what data the pipeline processed." % subtask.id)

        pipeline_map = siplib.PipelineMap(
                name=subtask.task_blueprint.name,
                version='unknown',  # todo from subtask.specifications_doc? from feedback (we have feedback and storagewriter versions there, not pipeline version or sth)?
                sourcedata_identifiers=sourcedata_identifiers,
                process_map=process_map)

        if subtask.specifications_template.name == "pipeline control":  #  todo: re-evaluate this because schema name might change
            spec = add_defaults_to_json_object_for_schema(subtask.specifications_doc, subtask.specifications_template.schema)
            pipeline = siplib.AveragingPipeline(  # <-- this is what we need for UC1
                pipeline_map,
                numberofcorrelateddataproducts=get_number_of_dataproducts_of_type(subtask, Dataformat.Choices.MEASUREMENTSET.value),
                frequencyintegrationstep=spec['demixer']['frequency_steps'] if spec['demixer']['enabled'] else 1,
                timeintegrationstep=spec['demixer']['time_steps'] if spec['demixer']['enabled'] else 1,
                flagautocorrelations=spec['preflagger1']['enabled'] and spec['preflagger1']['corrtype'] == 'auto',
                demixing=spec['demixer']['enabled'] and (spec['demixer']['demix_always'] or spec['demixer']['demix_if_needed'])
            )
        # todo: distinguish and create other pipeline types. Probably most of these can be filled in over time as needed,
        #  but they are not required for UC1. Here are stubs to start from for the other types the LTA supports:
        # elif subtask.specifications_template.name == "<???>": # todo
        #     pipeline = siplib.ImagingPipeline(pipeline_map,
        #                                imagerintegrationtime=10,  # todo
        #                                imagerintegrationtime_unit="ms",  #todo
        #                                numberofmajorcycles=5,  #todo
        #                                numberofinstrumentmodels=5,  #todo
        #                                numberofcorrelateddataproducts=1,  #todo
        #                                numberofskyimages=1, #todo
        #                                frequencyintegrationstep=None, #todo
        #                                timeintegrationstep=None, #todo
        #                                skymodeldatabase=None,  #todo
        #                                demixing=None)  #todo
        # elif subtask.specifications_template.name == "<???>":  # todo
        #     pipeline = siplib.CalibrationPipeline(
        #     pipeline_map,
        #     skymodeldatabase = "db", #todo
        #     numberofinstrumentmodels = 1,  #todo
        #     numberofcorrelateddataproducts = 1,  #todo
        #     frequencyintegrationstep = 1,  #todo
        #     timeintegrationstep = 1,  #todo
        #     flagautocorrelations = True,  #todo
        #     demixing = False)  #todo
        # elif subtask.specifications_template.name == "<???>":  # todo
        #     pipeline = siplib.PulsarPipeline(pipeline_map,
        #         pulsarselection="Pulsars in observation specs, file and brightest in SAP and TAB",   #todo
        #         pulsars=["J1234+67"],  #todo
        #         dosinglepulseanalysis=False,  #todo
        #         convertRawTo8bit=True,  #todo
        #         subintegrationlength=10,  #todo
        #         subintegrationlength_unit='ns',  #todo
        #         skiprfiexcision=False,  #todo
        #         skipdatafolding=False,  #todo
        #         skipoptimizepulsarprofile=True,  #todo
        #         skipconvertrawintofoldedpsrfits=False,  #todo
        #         runrotationalradiotransientsanalysis=True,   #todo
        #         skipdynamicspectrum=False,  #todo
        #         skipprefold=True)  #todo
        else:
            logger.warning('Could not identify LTA pipeline flavor for subtask=%s, template name=%s). Adding as SimplePipeline to SIP.'
                           % (subtask, subtask.specifications_template.name))
            pipeline = siplib.SimplePipeline(pipeline_map)
        return pipeline
    else:
        # todo: also cover, if needed: COPY, INSPECTION, QA_FILES, QA_PLOTS, DELETION, MANUAL, OTHER
        #  could be handled as UnspecifiedProcess, or require modeling in the SIP as well?
        raise TMSSException("The Subtask type %s cannot be represented in a SIP" % subtask.specifications_template.type.value)


def create_sip_representation_for_dataproduct(dataproduct: Dataproduct):
    """
    Extract info from the TMSS dataproduct and creates a SIP object representation
    :param dataproduct:
    :return: One of the siplib dataproduct object flavors.
    """

    # Make sure feedback is not empty
    if 'percentage_written' not in dataproduct.feedback_doc:
        raise ValueError("The feedback_doc of dataproduct id=%s is incomplete. Has feedback of the producing subtask been processed?" % dataproduct.id)

    # Note: this is for the type property present on all dataproduct flavors, dataproduct classes are
    #  differentiated in addition to that below
    type_map = {Datatype.Choices.VISIBILITIES.value: constants.DATAPRODUCTTYPE_CORRELATOR_DATA,
                Datatype.Choices.TIME_SERIES.value: constants.DATAPRODUCTTYPE_BEAM_FORMED_DATA,
                Datatype.Choices.INSTRUMENT_MODEL.value : constants.DATAPRODUCTTYPE_INSTRUMENT_MODEL,
                Datatype.Choices.IMAGE.value : constants.DATAPRODUCTTYPE_SKY_IMAGE,
                Datatype.Choices.QUALITY.value : constants.DATAPRODUCTTYPE_UNKNOWN}

    # Note: this is for the fileformat property present on all dataproduct flavors,
    #  dataproduct classes are differentiated in addition to that below
    fileformat_map = {Dataformat.Choices.MEASUREMENTSET.value: constants.FILEFORMATTYPE_AIPS___CASA,
                      Dataformat.Choices.BEAMFORMED.value: constants.FILEFORMATTYPE_HDF5,
                      Dataformat.Choices.QA_HDF5.value: constants.FILEFORMATTYPE_HDF5,
                      Dataformat.Choices.QA_PLOTS.value: constants.FILEFORMATTYPE_UNDOCUMENTED}

    storage_writer_map = {"dysco": constants.STORAGEWRITERTYPE_DYSCOSTORAGEMANAGER,
                          "unknown": constants.STORAGEWRITERTYPE_UNKNOWN,
                          "standard": constants.STORAGEWRITERTYPE_LOFARSTORAGEMANAGER,
                          "lofarstman": constants.STORAGEWRITERTYPE_LOFARSTORAGEMANAGER,}

    try:
        dataproduct_type = type_map[dataproduct.datatype.value]
    except Exception as err:
        dataproduct_type = constants.DATAPRODUCTTYPE_UNKNOWN
        logger.warning("Could not determine the type of dataproduct id %s (%s). Falling back to %s" % (dataproduct.id, err, dataproduct_type))

    try:
        dataproduct_fileformat = fileformat_map[dataproduct.producer.subtask.task_blueprint.consumed_by.first().dataformat.value] # todo same as with type? Why is this not with the data? Why is this so different from the LTA datamodel?
    except Exception as err:
        dataproduct_fileformat = constants.FILEFORMATTYPE_UNDOCUMENTED
        logger.warning("Could not determine the type of dataproduct id %s (%s). Falling back to %s" % (dataproduct.id, err, dataproduct_fileformat))

    dataproduct_map = siplib.DataProductMap(type=dataproduct_type,
                                            identifier=get_siplib_identifier(dataproduct.global_identifier, "Dataproduct %s" % dataproduct.id),
                                            size=dataproduct.size if dataproduct.size else 0,
                                            filename=dataproduct.filename,
                                            fileformat=dataproduct_fileformat,
                                            storage_writer=storage_writer_map[dataproduct.feedback_doc["samples"]["writer"] if 'samples' in dataproduct.feedback_doc else 'unknown'], # todo: verify we can use the feedback_doc here and remove the old method | storage_writer_map[dataproduct.producer.subtask.task_blueprint.specifications_doc.get("storagemanager", 'unknown')],
                                            storage_writer_version=dataproduct.feedback_doc["samples"]["writer_version"] if 'samples' in dataproduct.feedback_doc else 'unknown',
                                            process_identifier=get_siplib_identifier(dataproduct.producer.subtask.global_identifier, "Producer Subtask %s" % dataproduct.producer.subtask.id))

    if dataproduct.dataformat.value == Dataformat.Choices.MEASUREMENTSET.value:  # <- This is the only one we currently need for UC1
        sip_dataproduct = siplib.CorrelatedDataProduct(
            dataproduct_map,
            subarraypointing_identifier=get_siplib_identifier(dataproduct.sap.global_identifier, "SAP %s" % dataproduct.sap.id),
            subband=dataproduct.feedback_doc['frequency']['subbands'][0],
            starttime=dataproduct.feedback_doc['time']['start_time'],
            duration=isodate.duration_isoformat(datetime.timedelta(seconds=round(dataproduct.feedback_doc['time']['duration']))),
            integrationinterval=dataproduct.feedback_doc['time']['sample_width'],
            integrationintervalunit="s",
            central_frequency=dataproduct.feedback_doc['frequency']['central_frequencies'][0],
            central_frequencyunit="Hz",
            channelwidth_frequency=dataproduct.feedback_doc['frequency']['channel_width'],
            channelwidth_frequencyunit="Hz",
            channelspersubband=dataproduct.feedback_doc['frequency']['channels_per_subband'],
            stationsubband=0  # not correct ;)    (see metadata recipe CEP/Pipeline/recipes/sip/helpers/metadata.py)
        )
    # todo: distinguish and create other dataproduct types. Probably most of these can be filled in over time as needed,
    #  but they are not required for UC1. Here are stubs to start from for the other types the LTA supports:
    # elif dataproduct.dataformat.value == Dataformat.Choices.<???>.value: # todo
    #     sip_dataproduct = siplib.SkyModelDataProduct(dataproduct_map)
    # elif dataproduct.dataformat.value == Dataformat.Choices. <??? >.value:  # todo
    #     sip_dataproduct = siplip.GenericDataProduct(dataproduct_map)
    # elif dataproduct.dataformat.value == Dataformat.Choices. <??? >.value:  # todo
    #     sip_dataproduct = siplip.UnspecifiedDataProduct(dataproduct_map)
    # elif dataproduct.dataformat.value == Dataformat.Choices. <??? >.value:  # todo
    #     sip_dataproduct = siplip.PixelMapDataProduct(dataproduct_map,
    #                                                  numberofaxes=5, # todo
    #                                                  coordinates=[siplib.SpectralCoordinate(  # todo
    #                                                      quantity_type="Frequency",  # todo
    #                                                      quantity_value=20.0,  #todo
    #                                                      axis=siplib.LinearAxis(  #todo
    #                                                          number=5,  #todo
    #                                                          name="bla",  #todo
    #                                                          units="parsec",  #todo
    #                                                          length=5,  #todo
    #                                                          increment=5,  #todo
    #                                                          referencepixel=7.5,  #todo
    #                                                          referencevalue=7.4))])  #todo
    # elif dataproduct.dataformat.value == Dataformat.Choices. <??? >.value:  # todo
    #     sip_dataproduct = siplib.SkyImageDataProduct(
    #         dataproduct_map,  # todo
    #         numberofaxes=2,  # todo
    #         coordinates=[  # todo
    #             siplib.SpectralCoordinate(  # todo
    #                 quantity_type=constants.SPECTRALQUANTITYTYPE_VELOCITYAPPRADIAL,  # todo
    #                 quantity_value=20.0,  # todo
    #                 axis=siplib.LinearAxis(   # todo
    #                     number=5,  # todo
    #                     name="bla",  # todo
    #                     units="unit",  # todo
    #                     length=5,  # todo
    #                     increment=5,  # todo
    #                     referencepixel=7.5,  # todo
    #                     referencevalue=7.4)),  # todo
    #             siplib.DirectionCoordinate(  # todo
    #                 linearaxis_a=siplib.LinearAxis(  # todo
    #                     number=5,  # todo
    #                     name="bla",  # todo
    #                     units="unit",  # todo
    #                     length=5,  # todo
    #                     increment=5,  # todo
    #                     referencepixel=7.5,  # todo
    #                     referencevalue=7.4),  # todo
    #                 linearaxis_b=siplib.LinearAxis(  # todo
    #                     number=5,  # todo
    #                     name="blb",  # todo
    #                     units="unit",  # todo
    #                     length=5,  # todo
    #                     increment=5,  # todo
    #                     referencepixel=7.5,  # todo
    #                     referencevalue=7.4),  # todo
    #                 pc0_0=0.0,  # todo
    #                 pc0_1=0.1,  # todo
    #                 pc1_0=1.0,  # todo
    #                 pc1_1=1.1,  # todo
    #                 equinox=constants.EQUINOXTYPE_JUPITER,  # todo
    #                 radecsystem=constants.RADECSYSTEM_FK4_NO_E,  # todo
    #                 projection="rear",  # todo
    #                 projectionparameters=[1.0, 1.0, 1.0],  # todo
    #                 longitudepole_angle=1.0,  # todo
    #                 longitudepole_angleunit=constants.ANGLEUNIT_RADIANS,  # todo
    #                 latitudepole_angle=2.0,  # todo
    #                 latitudepole_angleunit=constants.ANGLEUNIT_ARCSEC,  # todo
    #             ),
    #             siplib.PolarizationCoordinate(  # todo
    #                 tabularaxis=siplib.TabularAxis(  # todo
    #                     number=5,  # todo
    #                     name="bla",  # todo
    #                     units="someunit",  # todo
    #                     length=5,  # todo
    #                 ),
    #                 polarizations=[constants.POLARIZATIONTYPE_LR, constants.POLARIZATIONTYPE_XRE]  # todo
    #             ),
    #         ],
    #         locationframe=constants.LOCATIONFRAME_LOCAL_GROUP,  # todo
    #         timeframe="timeframe",  # todo
    #         observationpointing=siplib.PointingRaDec(  # todo
    #             ra_angle=1.0,  # todo
    #             ra_angleunit=constants.ANGLEUNIT_DEGREES,  # todo
    #             dec_angle=42.0,  # todo
    #             dec_angleunit=constants.ANGLEUNIT_DEGREES,  # todo
    #             equinox=constants.EQUINOXTYPE_B1950  # todo
    #         ),
    #         restoringbeammajor_angle=1.0,  # todo
    #         restoringbeammajor_angleunit=constants.ANGLEUNIT_DEGREES,  # todo
    #         restoringbeamminor_angle=2.0,  # todo
    #         restoringbeamminor_angleunit=constants.ANGLEUNIT_DEGREES,  # todo
    #         rmsnoise=1.0  # todo
    #     )
    else:
        logger.warning('Could not identify corresponding SIP dataproduct flavor for dataproduct=%s, dataformat=%s. Adding as SimpleDataproduct to SIP.' % (dataproduct, dataproduct.dataformat))
        sip_dataproduct = siplib.SimpleDataProduct(dataproduct_map)
    return sip_dataproduct


def recursively_add_subtasks_to_sip(subtask: Subtask, sip: siplib.Sip):
    """
    add a subtask and the whole tree of its predecessor subtasks to the sip
    by traversing the subtask tree and repeatedly calling add_subbtask_to_sip().
    This will also add the input dataproducts of each subtask.
    :param subtask: the Subtask object to add to the sip
    :param sip: the Sip object to add to
    :return:
    """
    add_subbtask_and_input_dataproducts_to_sip(subtask, sip)
    for predecessor in subtask.predecessors.all():
        add_subbtask_and_input_dataproducts_to_sip(predecessor, sip)
        if subtask is not None:
            recursively_add_subtasks_to_sip(predecessor, sip)


def generate_sip_for_dataproduct(dataproduct: Dataproduct) -> siplib.Sip:
    """
    generate_sip_from_dataproduct
    :param dataproduct: The dataproduct to create the SIP for
    :return: The Sip object
    """

    # Create sip representation of the described/main dataproduct
    if dataproduct.producer.subtask.specifications_template.type.value == SubtaskType.Choices.INGEST.value:
        # the output dataproduct of an ingest subtask is a special case
        # generate the SIP for the corresponding input dataproduct
        ingest_input_dataproduct = dataproduct.producer.subtask.get_transformed_input_dataproduct(dataproduct)
        sip_dataproduct = create_sip_representation_for_dataproduct(ingest_input_dataproduct)

        # and enrich it SIP with the archive info of the output dataproduct
        if dataproduct.archive_info:
            sip_dataproduct.set_storageTicket(dataproduct.archive_info.storage_ticket)
            # enrich/overwrite the filename/size by the output dataproduct filename/size
            sip_dataproduct.set_fileName(dataproduct.filename)
            sip_dataproduct.set_size(dataproduct.size)

        if dataproduct.hashes:
            from django.core.exceptions import ObjectDoesNotExist
            try:
                sip_dataproduct.set_checksum_md5(dataproduct.hashes.get(algorithm=Algorithm.Choices.MD5.value).hash)
            except ObjectDoesNotExist:
                pass

            try:
                sip_dataproduct.set_checksum_adler32(dataproduct.hashes.get(algorithm=Algorithm.Choices.ADLER32.value).hash)
            except ObjectDoesNotExist:
                pass

        # for the rest of the sip generation, use the ingest_input_dataproduct
        dataproduct = ingest_input_dataproduct
    else:
        # normal case: a normal dataproduct generated by a normal observation/pipeline subtask
        sip_dataproduct = create_sip_representation_for_dataproduct(dataproduct)

    # Gather project details
    project = dataproduct.producer.subtask.task_blueprint.scheduling_unit_blueprint.draft.scheduling_set.project
    project_code = project.name
    project_primaryinvestigator = 'project_primaryinvestigator'
    project_contactauthor = 'project_contactauthor'
    project_description = 'project_description'

    # Create the Sip object
    sip = siplib.Sip(project_code=project_code,
                     project_primaryinvestigator=project_primaryinvestigator,
                     project_contactauthor=project_contactauthor,
                     project_description=project_description,
                     dataproduct=sip_dataproduct)

    subtask = dataproduct.producer.subtask
    recursively_add_subtasks_to_sip(subtask, sip)

    validator.check_consistency(sip)
    return sip

