"""
This module 'populate' defines methods to populate the database with predefined ('static') data,
according to the proposed Django way: https://docs.djangoproject.com/en/2.0/topics/migrations/#data-migrations

import this module in your empty migration step file, and add the following migration:

from ..populate import *

class Migration(migrations.Migration):

    dependencies = [ <the dependency is automatically inserted here> ]

    operations = [ migrations.RunPython(populate_choices) ]

"""

import logging
logger = logging.getLogger(__name__)

import inspect
import re
from datetime import timezone, datetime, date
from lofar.sas.tmss.tmss.tmssapp import models
from lofar.sas.tmss.tmss.tmssapp import viewsets
from lofar.sas.tmss.tmss.tmssapp.models.specification import *
from lofar.sas.tmss.tmss.tmssapp.models.scheduling import *
from lofar.sas.tmss.tmss.tmssapp.models.permissions import *
from lofar.sas.tmss.tmss.tmssapp.conversions import timestamps_and_stations_to_sun_rise_and_set, get_all_stations
from lofar.common import isTestEnvironment, isDevelopmentEnvironment
from concurrent.futures import ThreadPoolExecutor
from django.contrib.auth.models import User, Group, Permission
from django.contrib.contenttypes.models import ContentType
from django.db.utils import IntegrityError

working_dir = os.path.dirname(os.path.abspath(__file__))


def populate_choices(apps, schema_editor):
    '''
    populate each 'choice' table in the database with the 'static' list of 'choice'.Choices for
    each 'choice'type in Role, Datatype, Dataformat, CopyReason
    :return: None
    '''
    choice_classes = [Role, IOType, Datatype, Dataformat, CopyReason,
                      SubtaskState, SubtaskType, StationType, Algorithm, SchedulingRelationPlacement,
                      Flag, ProjectCategory, PeriodCategory, Quantity, TaskType, ProjectRole]

    # upload choices in parallel
    with ThreadPoolExecutor() as executor:
        executor.map(lambda choice_class: choice_class.objects.bulk_create([choice_class(value=x.value) for x in choice_class.Choices]),
                     choice_classes)

def populate_settings(apps, schema_editor):
    Setting.objects.create(name=Flag.objects.get(value='dynamic_scheduling_enabled'), value=False)

def populate_test_data():
    """
    Create a Test Schedule Set to be able to refer to when Scheduling Unit Draft is created from a
    scheduling unit json
    :return:
    """
    try:
        # only add (with  expensive setup time) example data when developing/testing and we're not unittesting
        if isTestEnvironment() or isDevelopmentEnvironment():
            from lofar.sas.tmss.tmss.exceptions import TMSSException
            from lofar.sas.tmss.test.tmss_test_data_django_models import SchedulingUnitDraft_test_data
            from lofar.sas.tmss.tmss.tmssapp.tasks import create_task_blueprints_and_subtasks_from_scheduling_unit_draft, create_task_blueprints_and_subtasks_and_schedule_subtasks_from_scheduling_unit_draft
            from lofar.sas.tmss.tmss.tmssapp.subtasks import schedule_subtask
            from lofar.common.json_utils import get_default_json_object_for_schema

            constraints_template = models.SchedulingConstraintsTemplate.objects.get(name="constraints")
            constraints_spec = get_default_json_object_for_schema(constraints_template.schema)

            uc1_strategy_template = models.SchedulingUnitObservingStrategyTemplate.objects.get(name="UC1 CTC+pipelines")
            simple_strategy_template = models.SchedulingUnitObservingStrategyTemplate.objects.get(name="Simple Observation")
            simple_beamforming_strategy_template = models.SchedulingUnitObservingStrategyTemplate.objects.get(name="Simple Beamforming Observation")

            projects = models.Project.objects.order_by('-priority_rank').all()
            for tmss_project in projects:
                if 'Commissioning' not in tmss_project.tags:
                    continue

                # for test purposes also create reservation objects from all reservation strategies
                for strategy_template in ReservationStrategyTemplate.objects.all():
                    reservation_spec = add_defaults_to_json_object_for_schema(strategy_template.template,
                                                                              strategy_template.reservation_template.schema)
                    reservation = Reservation.objects.create(name=strategy_template.name,
                                                             description=" %s created from reservation strategy" % strategy_template.description,
                                                             project=None,
                                                             specifications_template=strategy_template.reservation_template,
                                                             specifications_doc=reservation_spec,
                                                             start_time=datetime.now()+timedelta(days=1),
                                                             stop_time=None)
                    logger.info('created test reservation: %s', reservation.name)

                for scheduling_set in tmss_project.scheduling_sets.all():
                    for unit_nr in range(2):
                        for strategy_template in [simple_beamforming_strategy_template, uc1_strategy_template, simple_strategy_template]:
                            # the 'template' in the strategy_template is a predefined json-data blob which validates against the given scheduling_unit_template
                            # a user might 'upload' a partial json-data blob, so add all the known defaults
                            scheduling_unit_spec = add_defaults_to_json_object_for_schema(strategy_template.template, strategy_template.scheduling_unit_template.schema)

                            # limit target obs duration for demo data
                            if strategy_template == uc1_strategy_template:
                                scheduling_unit_spec['tasks']['Calibrator Observation 1']['specifications_doc']['duration'] = 2*60
                                scheduling_unit_spec['tasks']['Target Observation']['specifications_doc']['duration'] = 2*3600
                                scheduling_unit_spec['tasks']['Calibrator Observation 2']['specifications_doc']['duration'] = 2*60
                            elif strategy_template == simple_strategy_template:
                                scheduling_unit_spec['tasks']['Observation']['specifications_doc']['duration'] = 5*60

                            # set some constraints, so the dynamic scheduler has something to chew on.
                            # DISABLED for now, because the 'daily' constraint solver is not ready yet.
                            # constraints_spec['daily']['require_day'] = unit_nr%2==0
                            # constraints_spec['daily']['require_night'] = unit_nr%2==1
                            # constraints_spec['daily']['avoid_twilight'] = unit_nr%4>1

                            # add the scheduling_unit_doc to a new SchedulingUnitDraft instance, and were ready to use it!
                            scheduling_unit_draft = models.SchedulingUnitDraft.objects.create(name="%s %s %0d" % ('UC1' if strategy_template==uc1_strategy_template else 'Obs', tmss_project.name, unit_nr+1),
                                                                                              scheduling_set=scheduling_set,
                                                                                              description="Test scheduling unit",
                                                                                              requirements_template=strategy_template.scheduling_unit_template,
                                                                                              requirements_doc=scheduling_unit_spec,
                                                                                              observation_strategy_template=strategy_template,
                                                                                              scheduling_constraints_doc=constraints_spec,
                                                                                              scheduling_constraints_template=constraints_template)

                            logger.info('created test scheduling_unit_draft: %s', scheduling_unit_draft.name)

                            try:
                                create_task_blueprints_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)
                            except TMSSException as e:
                                logger.exception(e)
                        return

    except ImportError:
        pass


def populate_cycles(apps, schema_editor):
    #  Cycle 0 deviates from any patterns
    cycle = models.Cycle.objects.create(name="Cycle 00",
                                        description="Lofar Cycle 0",
                                        start=datetime(2013, 2, 11, 0, 0, 0, 0, tzinfo=timezone.utc),
                                        stop=datetime(2013, 11, 14, 0, 0, 0, 0, tzinfo=timezone.utc))

    models.CycleQuota.objects.bulk_create([models.CycleQuota(cycle=cycle,
                                                             resource_type=ResourceType.objects.get(
                                                                 name="LOFAR Observing Time"),
                                                             value=0.8 * cycle.duration.total_seconds()),
                                           # rough guess. 80% of total time available for observing
                                           models.CycleQuota(cycle=cycle,
                                                             resource_type=ResourceType.objects.get(
                                                                 name="CEP Processing Time"),
                                                             value=0.8 * cycle.duration.total_seconds()),
                                           models.CycleQuota(cycle=cycle,
                                                             resource_type=ResourceType.objects.get(name="LTA Storage"),
                                                             value=0),  # needs to be filled in by user (SOS)
                                           models.CycleQuota(cycle=cycle,
                                                             resource_type=ResourceType.objects.get(
                                                                 name="LOFAR Support Time"),
                                                             value=0),  # needs to be filled in by user (SOS)

                                           models.CycleQuota(cycle=cycle,
                                                             resource_type=ResourceType.objects.get(
                                                                 name="LOFAR Observing Time Commissioning"),
                                                             value=0.05 * cycle.duration.total_seconds()),
                                           # rough guess. 5% of total time available for observing
                                           models.CycleQuota(cycle=cycle,
                                                             resource_type=ResourceType.objects.get(
                                                                 name="LOFAR Observing Time prio A"),
                                                             value=0),  # needs to be filled in by user (SOS)
                                           models.CycleQuota(cycle=cycle,
                                                             resource_type=ResourceType.objects.get(
                                                                 name="LOFAR Observing Time prio B"),
                                                             value=0)  # needs to be filled in by user (SOS)
                                           ])

    #  Cycles 1-10 follow the same pattern
    for nr in range(1, 11):
        cycle = models.Cycle.objects.create(name="Cycle %02d" % nr,
                                            description="Lofar Cycle %s" % nr,
                                            start=datetime(2013+nr//2, 5 if nr%2==0 else 11, 15, 0, 0, 0, 0, tzinfo=timezone.utc),
                                            stop=datetime(2013+(nr+1)//2, 5 if nr%2==1 else 11, 14, 0, 0, 0, 0, tzinfo=timezone.utc))

        models.CycleQuota.objects.bulk_create([models.CycleQuota(cycle=cycle,
                                                                 resource_type=ResourceType.objects.get(name="LOFAR Observing Time"),
                                                                 value=0.8*cycle.duration.total_seconds()), # rough guess. 80% of total time available for observing
                                               models.CycleQuota(cycle=cycle,
                                                                 resource_type=ResourceType.objects.get(name="CEP Processing Time"),
                                                                 value=0.8*cycle.duration.total_seconds()),
                                               models.CycleQuota(cycle=cycle,
                                                                 resource_type=ResourceType.objects.get(name="LTA Storage"),
                                                                 value=0), # needs to be filled in by user (SOS)
                                               models.CycleQuota(cycle=cycle,
                                                                 resource_type=ResourceType.objects.get(name="LOFAR Support Time"),
                                                                 value=0),  # needs to be filled in by user (SOS)

                                               models.CycleQuota(cycle=cycle,
                                                                 resource_type=ResourceType.objects.get(name="LOFAR Observing Time Commissioning"),
                                                                 value=0.05*cycle.duration.total_seconds()), # rough guess. 5% of total time available for observing
                                               models.CycleQuota(cycle=cycle,
                                                                 resource_type=ResourceType.objects.get(name="LOFAR Observing Time prio A"),
                                                                 value=0), # needs to be filled in by user (SOS)
                                               models.CycleQuota(cycle=cycle,
                                                                 resource_type=ResourceType.objects.get(name="LOFAR Observing Time prio B"),
                                                                 value=0) # needs to be filled in by user (SOS)
                                               ])

    #  Cycle 11 deviates from any patterns
    cycle = models.Cycle.objects.create(name="Cycle 11",
                                        description="Lofar Cycle 11",
                                        start=datetime(2018, 11, 15, 0, 0, 0, 0, tzinfo=timezone.utc),
                                        stop=datetime(2019, 5, 31, 0, 0, 0, 0, tzinfo=timezone.utc))

    models.CycleQuota.objects.bulk_create([models.CycleQuota(cycle=cycle,
                                                             resource_type=ResourceType.objects.get(
                                                                 name="LOFAR Observing Time"),
                                                             value=0.8 * cycle.duration.total_seconds()),
                                           # rough guess. 80% of total time available for observing
                                           models.CycleQuota(cycle=cycle,
                                                             resource_type=ResourceType.objects.get(
                                                                 name="CEP Processing Time"),
                                                             value=0.8 * cycle.duration.total_seconds()),
                                           models.CycleQuota(cycle=cycle,
                                                             resource_type=ResourceType.objects.get(
                                                                 name="LTA Storage"),
                                                             value=0),  # needs to be filled in by user (SOS)
                                           models.CycleQuota(cycle=cycle,
                                                             resource_type=ResourceType.objects.get(
                                                                 name="LOFAR Support Time"),
                                                             value=0),  # needs to be filled in by user (SOS)

                                           models.CycleQuota(cycle=cycle,
                                                             resource_type=ResourceType.objects.get(
                                                                 name="LOFAR Observing Time Commissioning"),
                                                             value=0.05 * cycle.duration.total_seconds()),
                                           # rough guess. 5% of total time available for observing
                                           models.CycleQuota(cycle=cycle,
                                                             resource_type=ResourceType.objects.get(
                                                                 name="LOFAR Observing Time prio A"),
                                                             value=0),  # needs to be filled in by user (SOS)
                                           models.CycleQuota(cycle=cycle,
                                                             resource_type=ResourceType.objects.get(
                                                                 name="LOFAR Observing Time prio B"),
                                                             value=0)  # needs to be filled in by user (SOS)
                                           ])

    #  Cycles 12-19 follow the same pattern
    for nr in range(12, 20):
        cycle = models.Cycle.objects.create(name="Cycle %02d" % nr,
                                            description="Lofar Cycle %s" % nr,
                                            start=datetime(2013 + nr // 2, 6 if nr % 2 == 0 else 12, 1, 0, 0, 0, 0,
                                                           tzinfo=timezone.utc),
                                            stop=datetime(2013 + (nr + 1) // 2, 5 if nr % 2 == 1 else 11,
                                                          30 if nr % 2 == 0 else 31, 0, 0,
                                                          0, 0, tzinfo=timezone.utc))

        models.CycleQuota.objects.bulk_create([models.CycleQuota(cycle=cycle,
                                                                 resource_type=ResourceType.objects.get(
                                                                     name="LOFAR Observing Time"),
                                                                 value=0.8 * cycle.duration.total_seconds()),
                                               # rough guess. 80% of total time available for observing
                                               models.CycleQuota(cycle=cycle,
                                                                 resource_type=ResourceType.objects.get(
                                                                     name="CEP Processing Time"),
                                                                 value=0.8 * cycle.duration.total_seconds()),
                                               models.CycleQuota(cycle=cycle,
                                                                 resource_type=ResourceType.objects.get(
                                                                     name="LTA Storage"),
                                                                 value=0),  # needs to be filled in by user (SOS)
                                               models.CycleQuota(cycle=cycle,
                                                                 resource_type=ResourceType.objects.get(
                                                                     name="LOFAR Support Time"),
                                                                 value=0),  # needs to be filled in by user (SOS)

                                               models.CycleQuota(cycle=cycle,
                                                                 resource_type=ResourceType.objects.get(
                                                                     name="LOFAR Observing Time Commissioning"),
                                                                 value=0.05 * cycle.duration.total_seconds()),
                                               # rough guess. 5% of total time available for observing
                                               models.CycleQuota(cycle=cycle,
                                                                 resource_type=ResourceType.objects.get(
                                                                     name="LOFAR Observing Time prio A"),
                                                                 value=0),  # needs to be filled in by user (SOS)
                                               models.CycleQuota(cycle=cycle,
                                                                 resource_type=ResourceType.objects.get(
                                                                     name="LOFAR Observing Time prio B"),
                                                                 value=0)  # needs to be filled in by user (SOS)
                                               ])


def populate_projects(apps, schema_editor):
    from lofar.sas.tmss.test.tmss_test_data_django_models import SchedulingSet_test_data

    for name, rank in (("high", 3), ("normal", 2), ("low", 1)):
        tmss_project = models.Project.objects.create(name=name,
                                                 description="Project for all TMSS tests and commissioning (%s priority)" % (name,),
                                                 priority_rank=rank,
                                                 can_trigger=False,
                                                 private_data=True,
                                                 expert=True,
                                                 filler=False)
        tmss_project.tags = ["Commissioning"]
        tmss_project.cycles.set([models.Cycle.objects.get(name="Cycle 14")])
        tmss_project.save()

        # for convenience, create a schedulingset for each project
        models.SchedulingSet.objects.create(**SchedulingSet_test_data(name="Test Scheduling Set", project=tmss_project))

        project_quota = ProjectQuota.objects.create(project=tmss_project, value=1e12, resource_type=ResourceType.objects.get(name="LTA Storage"))
        sara_fs = Filesystem.objects.get(name="Lofar Storage (SARA)")
        models.ProjectQuotaArchiveLocation.objects.create(project_quota=project_quota, archive_location=sara_fs)


def populate_resources(apps, schema_editor):
    bytes_q = Quantity.objects.get(value=Quantity.Choices.BYTES.value)
    time_q = Quantity.objects.get(value=Quantity.Choices.TIME.value)
    number_q = Quantity.objects.get(value=Quantity.Choices.NUMBER.value)

    ResourceType.objects.bulk_create([ResourceType(name="LTA Storage", description="Amount of storage in the LTA (in bytes)", quantity=bytes_q),
                                      ResourceType(name="CEP Storage", description="Amount of storage on the CEP processing cluster (in bytes)", quantity=bytes_q),
                                      ResourceType(name="CEP Processing Time", description="Processing time on the CEP processing cluster (in seconds)", quantity=time_q),
                                      ResourceType(name="LOFAR Observing Time", description="Observing time (in seconds)", quantity=time_q),
                                      ResourceType(name="LOFAR Observing Time prio A", description="Observing time with priority A (in seconds)", quantity=time_q),
                                      ResourceType(name="LOFAR Observing Time prio B", description="Observing time with priority B (in seconds)", quantity=time_q),
                                      ResourceType(name="LOFAR Observing Time Commissioning", description="Observing time for Commissioning/DDT (in seconds)", quantity=time_q),
                                      ResourceType(name="LOFAR Support Time", description="Support time by human (in seconds)", quantity=time_q),
                                      ResourceType(name="Number of triggers", description="Number of trigger events (as integer)", quantity=number_q) ])


def populate_misc(apps, schema_editor):
    cluster = Cluster.objects.create(name="CEP4", location="CIT", archive_site=False)
    fs = Filesystem.objects.create(name="LustreFS", cluster=cluster, capacity=3.6e15)

    sara_cluster = Cluster.objects.create(name="SARA", location="SARA", archive_site=True)
    juelich_cluster = Cluster.objects.create(name="Jülich", location="Jülich", archive_site=True)
    poznan_cluster = Cluster.objects.create(name="Poznan", location="Poznan", archive_site=True)

    sara_fs = Filesystem.objects.create(name="Lofar Storage (SARA)", cluster=sara_cluster, capacity=3.6e15,
                                        directory="srm://srm.grid.sara.nl:8443/pnfs/grid.sara.nl/data/lofar/ops/projects/")
    sara_test_fs = Filesystem.objects.create(name="Lofar Test Storage (SARA)", cluster=sara_cluster, capacity=3.6e15,
                                             directory="srm://srm.grid.sara.nl:8443/pnfs/grid.sara.nl/data/lofar/ops/test/projects/")
    sara_user_fs = Filesystem.objects.create(name="Lofar User Disk Storage (SARA)", cluster=sara_cluster, capacity=3.6e15,
                                             directory="srm://srm.grid.sara.nl/pnfs/grid.sara.nl/data/lofar/user/disk/projects/")
    juelich_fs = Filesystem.objects.create(name="Lofar Storage (Jülich)", cluster=juelich_cluster, capacity=3.6e15,
                                           directory="srm://lofar-srm.fz-juelich.de:8443/pnfs/fz-juelich.de/data/lofar/ops/projects/")
    pozname_fs = Filesystem.objects.create(name="Lofar Storage (Poznan)", cluster=poznan_cluster, capacity=3.6e15,
                                           directory="srm://lta-head.lofar.psnc.pl:8443/lofar/ops/projects/")


def populate_connectors():
    # the TaskConnectorType's define how the Task[Draft/Blueprint] *can* be connected.

    # NOTE: This is an explicit list of each possible link between tasks. This model suffices
    # until the number of connectors throw too large. By then, we could consider introducing
    # wild cards, like output_of=NULL meaning "any".
    logger.info("POPULATING CONNECTORS")

    # calibrator observation
    TaskConnectorType.objects.create(role=Role.objects.get(value=Role.Choices.CORRELATOR.value),
                                 datatype=Datatype.objects.get(value=Datatype.Choices.VISIBILITIES.value),
                                 task_template=TaskTemplate.objects.get(name='calibrator observation'),
                                 iotype=IOType.objects.get(value=IOType.Choices.OUTPUT.value))

    # target observation
    TaskConnectorType.objects.create(role=Role.objects.get(value=Role.Choices.CORRELATOR.value),
                                 datatype=Datatype.objects.get(value=Datatype.Choices.VISIBILITIES.value),
                                 task_template=TaskTemplate.objects.get(name='target observation'),
                                 iotype=IOType.objects.get(value=IOType.Choices.OUTPUT.value))

    # preprocessing pipeline
    TaskConnectorType.objects.create(role=Role.objects.get(value=Role.Choices.ANY.value),
                                 datatype=Datatype.objects.get(value=Datatype.Choices.VISIBILITIES.value),
                                 task_template=TaskTemplate.objects.get(name='preprocessing pipeline'),
                                 iotype=IOType.objects.get(value=IOType.Choices.INPUT.value))

    TaskConnectorType.objects.create(role=Role.objects.get(value=Role.Choices.ANY.value),
                                 datatype=Datatype.objects.get(value=Datatype.Choices.VISIBILITIES.value),
                                 task_template=TaskTemplate.objects.get(name='preprocessing pipeline'),
                                 iotype=IOType.objects.get(value=IOType.Choices.OUTPUT.value))

    # ingest
    TaskConnectorType.objects.create(role=Role.objects.get(value=Role.Choices.ANY.value),
                                 datatype=Datatype.objects.get(value=Datatype.Choices.VISIBILITIES.value),
                                 task_template=TaskTemplate.objects.get(name='ingest'),
                                 iotype=IOType.objects.get(value=IOType.Choices.INPUT.value))

    TaskConnectorType.objects.create(role=Role.objects.get(value=Role.Choices.ANY.value),
                                 datatype=Datatype.objects.get(value=Datatype.Choices.TIME_SERIES.value),
                                 task_template=TaskTemplate.objects.get(name='ingest'),
                                 iotype=IOType.objects.get(value=IOType.Choices.INPUT.value))


def populate_permissions():
    logger.info('Populating permissions')

    populate_project_permissions()
    populate_system_permissions()
    populate_system_roles()
    populate_system_test_users()


def populate_project_permissions():

    # For each viewset and for each extra action create a project permission entry.
    for name, obj in inspect.getmembers(viewsets):
        if inspect.isclass(obj):
            try:
                permission_name = obj.serializer_class.Meta.model.__name__.lower()
                logger.info('creating project permission %s' % permission_name)
                try:
                    ProjectPermission.objects.create(name=permission_name)
                except IntegrityError as e:
                    logger.debug('Skipping project permission creation for obj=%s: %s' % (obj, e))
                extra_actions = obj.get_extra_actions()
                if extra_actions:
                    for action in extra_actions:
                        action_permission_name = '%s-%s' % (permission_name, action.__name__)
                        logger.info('creating project permission %s' % action_permission_name)
                        try:
                            ProjectPermission.objects.create(name=action_permission_name)
                        except IntegrityError as e:
                            logger.debug('Skipping project permission creation for obj=%s: %s' % (obj, e))

            except Exception as e:
                logger.debug('Skipping project permission creation for obj=%s: %s' % (obj, e))


    # Project
    perm = ProjectPermission.objects.get(name='project')
    perm.GET.set([ProjectRole.objects.get(value='pi')])
    perm.GET.set([ProjectRole.objects.get(value='co_i')])
    perm.GET.set([ProjectRole.objects.get(value='contact_author')])
    perm.GET.set([ProjectRole.objects.get(value='shared_support_user')])
    perm.GET.set([ProjectRole.objects.get(value='friend_of_project')])
    perm.PATCH.set([ProjectRole.objects.get(value='friend_of_project')])
    perm.save()

    # Subtask

    # Subtask-schedule
    perm = ProjectPermission.objects.get(name='subtask-schedule')
    perm.GET.set([ProjectRole.objects.get(value='friend_of_project')])
    perm.save()

    # SchedulingUnitDraft
    perm = ProjectPermission.objects.get(name='schedulingunitdraft')
    perm.GET.set([ProjectRole.objects.get(value='shared_support_user')])
    perm.GET.set([ProjectRole.objects.get(value='friend_of_project')])
    perm.GET.set([ProjectRole.objects.get(value='contact_author')])
    perm.POST.set([ProjectRole.objects.get(value='shared_support_user')])
    perm.POST.set([ProjectRole.objects.get(value='friend_of_project')])
    perm.save()

    # SchedulingUnitBlueprint
    perm = ProjectPermission.objects.get(name='schedulingunitblueprint')
    perm.GET.set([ProjectRole.objects.get(value='shared_support_user')])
    perm.GET.set([ProjectRole.objects.get(value='friend_of_project')])
    perm.GET.set([ProjectRole.objects.get(value='contact_author')])
    perm.POST.set([ProjectRole.objects.get(value='shared_support_user')])   # "Let's try, we may want to revoke this later and review"
    perm.POST.set([ProjectRole.objects.get(value='friend_of_project')])
    perm.save()


def populate_system_permissions():
    # For each viewset create custom permissions for extra actions.
    for name, obj in inspect.getmembers(viewsets):
        if inspect.isclass(obj):
            try:
                ct = ContentType.objects.get_for_model(obj.serializer_class.Meta.model)
                extra_actions = obj.get_extra_actions()
                if extra_actions:
                    for action in extra_actions:
                        codename = ("%s_%s" % (action.__name__, obj.serializer_class.Meta.model.__name__)).lower()
                        name = f'Can {action.__name__} {obj.serializer_class.Meta.model.__name__.lower()}'
                        Permission.objects.create(codename=codename, name=name, content_type=ct)
            except:
                pass


def populate_system_roles():
    to_observer_group = Group.objects.create(name='TO observer')
    sdco_support_group = Group.objects.create(name='SDCO support')
    tmss_maintainer_group = Group.objects.create(name='TMSS Maintainer')
    tmss_admin_group = Group.objects.create(name='TMSS Admin')
    to_maintenance_group = Group.objects.create(name='TO maintenance')
    to_user_group = Group.objects.create(name='TO user')
    scientist_group = Group.objects.create(name='Scientist')
    e_scientist_group = Group.objects.create(name='Scientist (Expert)')
    guest_group = Group.objects.create(name='Guest')
    lta_user_group = Group.objects.create(name='LTA User')

    assign_system_permissions()


def assign_system_permissions():
    '''
    Assign system permission to each system role, accordingly.
    '''

    # Get system roles
    to_observer_group = Group.objects.get(name='TO observer')
    sdco_support_group = Group.objects.get(name='SDCO support')
    tmss_maintainer_group = Group.objects.get(name='TMSS Maintainer')
    tmss_admin_group = Group.objects.get(name='TMSS Admin')
    to_maintenance_group = Group.objects.get(name='TO maintenance')
    to_user_group = Group.objects.get(name='TO user')
    scientist_group = Group.objects.get(name='Scientist')
    e_scientist_group = Group.objects.get(name='Scientist (Expert)')
    guest_group = Group.objects.get(name='Guest')
    lta_user_group = Group.objects.get(name='LTA User')

    # Existing tests rely on this # todo: adapt tests to match actual permissions
    scientist_group.permissions.add(Permission.objects.get(codename='view_cycle'))
    scientist_group.permissions.add(Permission.objects.get(codename='view_project'))

    # Subtask model permissions
    ct = ContentType.objects.get(model='subtask')
    perm = Permission.objects.get(codename='view_subtask')
    to_observer_group.permissions.add(perm)
    sdco_support_group.permissions.add(perm)
    tmss_maintainer_group.permissions.add(perm)
    tmss_admin_group.permissions.add(perm)
    to_observer_group.permissions.add(Permission.objects.get(codename='schedule_subtask'))
    sdco_support_group.permissions.add(Permission.objects.get(codename='schedule_subtask'))
    tmss_maintainer_group.permissions.add(Permission.objects.get(codename='schedule_subtask'))
    tmss_admin_group.permissions.add(Permission.objects.get(codename='schedule_subtask'))
    # TODO: This is needed for testing atm. The table does not specify which roles should have this permission for now.
    for subtask_perm in ['get_progress_subtask', 'input_dataproducts_subtask', 'output_dataproducts_subtask',
                         'parset_subtask', 'predecessors_subtask', 'process_feedback_and_set_to_finished_if_complete_subtask',
                         'reprocess_raw_feedback_for_subtask_and_set_to_finished_if_complete_subtask', 'state_log_subtask',
                         'successors_subtask', 'task_log_subtask', 'transformed_output_dataproduct_subtask', 'unschedule_subtask' ]:
        to_observer_group.permissions.add(Permission.objects.get(codename=subtask_perm))

    # Template models permissions
    for name, obj in inspect.getmembers(models):
        if inspect.isclass(obj) and issubclass(obj, Template) and obj is not Template:
            ct = ContentType.objects.get(model=name.lower())
            perm_name = ''.join('_'.join(re.findall(r'[A-Z](?:[a-z]+|[A-Z]*(?=[A-Z]|$))', name)).lower())
            perm = Permission.objects.get(codename=f'add_{name.lower()}')
            to_observer_group.permissions.add(perm)
            sdco_support_group.permissions.add(perm)
            tmss_maintainer_group.permissions.add(perm)
            tmss_admin_group.permissions.add(perm)
            e_scientist_group.permissions.add(perm)
            perm = Permission.objects.get(codename=f'view_{name.lower()}')
            to_observer_group.permissions.add(perm)
            sdco_support_group.permissions.add(perm)
            tmss_maintainer_group.permissions.add(perm)
            tmss_admin_group.permissions.add(perm)
            e_scientist_group.permissions.add(perm)
            perm = Permission.objects.get(codename=f'change_{name.lower()}')
            to_observer_group.permissions.add(perm)
            sdco_support_group.permissions.add(perm)
            tmss_maintainer_group.permissions.add(perm)
            tmss_admin_group.permissions.add(perm)
            e_scientist_group.permissions.add(perm)

    # SchedulingUnit models permissions
    for template in ['schedulingunitdraft', 'schedulingunitblueprint']:
        ct = ContentType.objects.get(model=template)
        perm = Permission.objects.get(codename=f'add_{template}')
        to_observer_group.permissions.add(perm)
        sdco_support_group.permissions.add(perm)
        tmss_maintainer_group.permissions.add(perm)
        tmss_admin_group.permissions.add(perm)
        to_user_group.permissions.add(perm)
        e_scientist_group.permissions.add(perm) if template is 'schedulingunitdraft' else None
        perm = Permission.objects.get(codename=f'view_{template}')
        to_observer_group.permissions.add(perm)
        sdco_support_group.permissions.add(perm)
        tmss_maintainer_group.permissions.add(perm)
        tmss_admin_group.permissions.add(perm)
        to_user_group.permissions.add(perm)

    # Project model permissions
    ct = ContentType.objects.get(model='project')
    perm = Permission.objects.get(codename='add_project')
    to_observer_group.permissions.add(perm)
    sdco_support_group.permissions.add(perm)
    tmss_maintainer_group.permissions.add(perm)
    tmss_admin_group.permissions.add(perm)
    perm = Permission.objects.get(codename='view_project')
    to_observer_group.permissions.add(perm)
    sdco_support_group.permissions.add(perm)
    tmss_maintainer_group.permissions.add(perm)
    tmss_admin_group.permissions.add(perm)
    to_user_group.permissions.add(perm)
    perm = Permission.objects.get(codename='change_project')
    to_observer_group.permissions.add(perm)
    sdco_support_group.permissions.add(perm)
    tmss_maintainer_group.permissions.add(perm)
    tmss_admin_group.permissions.add(perm)

    # User model permissions
    ct = ContentType.objects.get(model='user')
    perm = Permission.objects.get(codename='add_user')
    to_observer_group.permissions.add(perm)
    sdco_support_group.permissions.add(perm)
    tmss_maintainer_group.permissions.add(perm)
    tmss_admin_group.permissions.add(perm)


def populate_system_test_users():
    # TODO: Set proper credentials (passwords at least).
    to_observer_user, _ = User.objects.get_or_create(username='to_observer', password='to_observer')
    to_observer_user.groups.add(Group.objects.get(name='TO observer'))
    sdco_support_user, _ = User.objects.get_or_create(username='sdco_support', password='sdco_support')
    sdco_support_user.groups.add(Group.objects.get(name='SDCO support'))
    tmss_maintainer_user, _ = User.objects.get_or_create(username='tmss_maintainer', password='tmss_maintainer')
    tmss_maintainer_user.groups.add(Group.objects.get(name='TMSS Maintainer'))
    tmss_admin_user, _ = User.objects.get_or_create(username='tmss_admin', password='tmss_admin')
    tmss_admin_user.groups.add(Group.objects.get(name='TMSS Admin'))
    to_maintenance_user, _ = User.objects.get_or_create(username='to_maintenance', password='to_maintenance')
    to_maintenance_user.groups.add(Group.objects.get(name='TO maintenance'))
    to_user, _ = User.objects.get_or_create(username='to_user', password='to_user')
    to_user.groups.add(Group.objects.get(name='TO user'))
    scientist_user, _ = User.objects.get_or_create(username='scientist', password='scientist')
    scientist_user.groups.add(Group.objects.get(name='Scientist'))
    e_scientist_user, _ = User.objects.get_or_create(username='e_scientist', password='e_scientist')
    e_scientist_user.groups.add(Group.objects.get(name='Scientist (Expert)'))
    guest_user, _ = User.objects.get_or_create(username='guest', password='guest')
    guest_user.groups.add(Group.objects.get(name='Guest'))
    lta_user, _ = User.objects.get_or_create(username='lta_user', password='lta_user')
    lta_user.groups.add(Group.objects.get(name='LTA User'))


def populate_sunrise_and_sunset_for_all_stations(nbr_days=3, start_date=date.today()):
    """
    Populate station timeline data of all stations for given number of days the starting at given date
    Note: If data is not in database yet, it will take about 6 seconds to calculate it for all (51) stations
    """
    starttime_for_logging = datetime.utcnow()
    logger.info("Populate sunrise and sunset for ALL known stations from %s up to %d days" % (start_date, nbr_days))
    lst_timestamps = []
    for i in range(0, nbr_days):
        dt = datetime.combine(start_date, datetime.min.time()) + timedelta(days=i)
        lst_timestamps.append(dt)

    timestamps_and_stations_to_sun_rise_and_set(tuple(lst_timestamps), tuple(get_all_stations()), create_when_not_found=True)
    logger.info("Populate sunrise and sunset done in %.1fs", (datetime.utcnow()-starttime_for_logging).total_seconds())
