"""
This file contains the serializers (for the elsewhere defined data models)
"""

from rest_framework import serializers
from .. import models
from .scheduling import SubtaskSerializer
from .common import FloatDurationField, RelationalHyperlinkedModelSerializer, AbstractTemplateSerializer, DynamicRelationalHyperlinkedModelSerializer
from .widgets import JSONEditorField
from django.contrib.auth.models import User

# This is required for keeping a user reference as ForeignKey in other models
# (I think so that the HyperlinkedModelSerializer can generate a URI)
class UserSerializer(serializers.Serializer):
    class Meta:
        model = User
        fields = '__all__'


class TagsSerializer(DynamicRelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.Tags
        fields = '__all__'


class CommonSchemaTemplateSerializer(AbstractTemplateSerializer):
    class Meta:
        model = models.CommonSchemaTemplate
        fields = '__all__'


class GeneratorTemplateSerializer(AbstractTemplateSerializer):
    class Meta:
        model = models.GeneratorTemplate
        fields = '__all__'


class DefaultGeneratorTemplateSerializer(DynamicRelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.DefaultGeneratorTemplate
        fields = '__all__'


class SchedulingUnitObservingStrategyTemplateSerializer(DynamicRelationalHyperlinkedModelSerializer):
    template = JSONEditorField(schema_source="scheduling_unit_template.schema")

    class Meta:
        model = models.SchedulingUnitObservingStrategyTemplate
        fields = '__all__'


class SchedulingUnitTemplateSerializer(AbstractTemplateSerializer):
    class Meta:
        model = models.SchedulingUnitTemplate
        fields = '__all__'


class DefaultSchedulingUnitTemplateSerializer(DynamicRelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.DefaultSchedulingUnitTemplate
        fields = '__all__'


class SchedulingConstraintsTemplateSerializer(AbstractTemplateSerializer):
    class Meta:
        model = models.SchedulingConstraintsTemplate
        fields = '__all__'


class DefaultSchedulingConstraintsTemplateSerializer(DynamicRelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.DefaultSchedulingConstraintsTemplate
        fields = '__all__'


class TaskTemplateSerializer(AbstractTemplateSerializer):
    class Meta:
        model = models.TaskTemplate
        fields = '__all__'


class DefaultTaskTemplateSerializer(DynamicRelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.DefaultTaskTemplate
        fields = '__all__'


class TaskRelationSelectionTemplateSerializer(AbstractTemplateSerializer):
    class Meta:
        model = models.TaskRelationSelectionTemplate
        fields = '__all__'


class DefaultTaskRelationSelectionTemplateSerializer(DynamicRelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.DefaultTaskRelationSelectionTemplate
        fields = '__all__'


class RoleSerializer(serializers.ModelSerializer):
    class Meta:
        model = models.Role
        fields = '__all__'

class IOTypeSerializer(serializers.ModelSerializer):
    class Meta:
        model = models.IOType
        fields = '__all__'

class SchedulingRelationPlacementSerializer(serializers.ModelSerializer):
    class Meta:
        model = models.SchedulingRelationPlacement
        fields = '__all__'


class DatatypeSerializer(serializers.ModelSerializer):
    class Meta:
        model = models.Datatype
        fields = '__all__'


class DataformatSerializer(DynamicRelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.Dataformat
        fields = '__all__'


class QuantitySerializer(DynamicRelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.Quantity
        fields = '__all__'


class CopyReasonSerializer(DynamicRelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.CopyReason
        fields = '__all__'


class TaskConnectorTypeSerializer(DynamicRelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.TaskConnectorType
        fields = '__all__'


class CycleSerializer(DynamicRelationalHyperlinkedModelSerializer):
    duration = FloatDurationField(read_only=True, help_text="Duration of the cycle [seconds]")

    class Meta:
        model = models.Cycle
        fields = '__all__'
        extra_fields = ['projects', 'name', 'duration', 'quota']
        expandable_fields = {
            'projects': ('lofar.sas.tmss.tmss.tmssapp.serializers.ProjectSerializer', {'many': True}),
            'quota': ('lofar.sas.tmss.tmss.tmssapp.serializers.CycleQuotaSerializer', {'many': True})
        }


class CycleQuotaSerializer(DynamicRelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.CycleQuota
        fields = '__all__'
        extra_fields = ['resource_type']

class ProjectSerializer(DynamicRelationalHyperlinkedModelSerializer):
#    scheduling_sets = serializers.PrimaryKeyRelatedField(source='scheduling_sets', read_only=True, many=True)

    class Meta:
        model = models.Project
        fields = '__all__'
        extra_fields = ['name','quota'] #, 'scheduling_sets']
        expandable_fields = {
            'cycles': ('lofar.sas.tmss.tmss.tmssapp.serializers.CycleSerializer', {'many': True}),
            'quota': ('lofar.sas.tmss.tmss.tmssapp.serializers.ProjectQuotaSerializer', {'many': True})
        }


class ProjectQuotaSerializer(DynamicRelationalHyperlinkedModelSerializer):
    project_quota_archive_location = serializers.HyperlinkedRelatedField('projectquotaarchivelocation-detail', source='*', read_only=True)

    class Meta:
        model = models.ProjectQuota
        fields = '__all__'
        extra_fields = ['resource_type', 'project_quota_archive_location']


class ProjectQuotaArchiveLocationSerializer(DynamicRelationalHyperlinkedModelSerializer):
    queryset = models.ProjectQuotaArchiveLocation.objects.all()

    # performance boost: select the related models in a single db call.
    queryset = queryset.select_related('project_quota', 'archive_location')

    class Meta:
        model = models.ProjectQuotaArchiveLocation
        fields = '__all__'
        extra_fields = ['archive_subdirectory', 'full_archive_uri']


class ResourceTypeSerializer(DynamicRelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.ResourceType
        fields = '__all__'
        extra_fields = ['name']


class FlagSerializer(DynamicRelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.Flag
        fields = '__all__'


class SettingSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        model = models.Setting
        fields = '__all__'


class ProjectCategorySerializer(DynamicRelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.ProjectCategory
        fields = '__all__'


class PeriodCategorySerializer(DynamicRelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.PeriodCategory
        fields = '__all__'


class SchedulingSetSerializer(DynamicRelationalHyperlinkedModelSerializer):
    generator_doc = JSONEditorField(schema_source="generator_template.schema")

    class Meta:
        model = models.SchedulingSet
        fields = '__all__'
        extra_fields = ['scheduling_unit_drafts']
        expandable_fields = {
            'generator_source': 'lofar.sas.tmss.tmss.tmssapp.serializers.SchedulingUnitDraftSerializer',
            'generator_template': 'lofar.sas.tmss.tmss.tmssapp.serializers.GeneratorTemplateSerializer',
            'project': 'lofar.sas.tmss.tmss.tmssapp.serializers.ProjectSerializer',
            'scheduling_unit_drafts': ('lofar.sas.tmss.tmss.tmssapp.serializers.SchedulingUnitDraftSerializer', {'many': True}),
        }


class SchedulingUnitDraftSerializer(DynamicRelationalHyperlinkedModelSerializer):
    requirements_doc = JSONEditorField(schema_source="requirements_template.schema")
    scheduling_constraints_doc = JSONEditorField(schema_source="scheduling_constraints_template.schema")
    duration = FloatDurationField(read_only=True)

    class Meta:
        model = models.SchedulingUnitDraft
        fields = '__all__'
        extra_fields = ['scheduling_unit_blueprints', 'task_drafts', 'duration']
        expandable_fields = {
            'observation_strategy_template': 'lofar.sas.tmss.tmss.tmssapp.serializers.SchedulingUnitObservingStrategyTemplateSerializer',
            'requirements_template': 'lofar.sas.tmss.tmss.tmssapp.serializers.SchedulingUnitTemplateSerializer',
            'scheduling_constraints_template': 'lofar.sas.tmss.tmss.tmssapp.serializers.SchedulingConstraintsTemplateSerializer',
            'scheduling_set': 'lofar.sas.tmss.tmss.tmssapp.serializers.SchedulingSetSerializer',
            'scheduling_unit_blueprints': ('lofar.sas.tmss.tmss.tmssapp.serializers.SchedulingUnitBlueprintSerializer', {'many': True}),
            'task_drafts': ('lofar.sas.tmss.tmss.tmssapp.serializers.TaskDraftSerializer', {'many': True})
        }


class SchedulingUnitDraftCopySerializer(SchedulingUnitDraftSerializer):
    class Meta(SchedulingUnitDraftSerializer.Meta):
       fields = ['copy_reason']
       extra_fields =['scheduling_set_id']
       read_only_fields = ['scheduling_unit_blueprints','task_drafts']


class SchedulingUnitDraftCopyFromSchedulingSetSerializer(SchedulingUnitDraftSerializer):
    class Meta(SchedulingUnitDraftSerializer.Meta):
       fields = ['copy_reason']
       read_only_fields = ['scheduling_unit_blueprints','task_drafts']


class SchedulingUnitBlueprintSerializer(DynamicRelationalHyperlinkedModelSerializer):
    requirements_doc = JSONEditorField(schema_source="requirements_template.schema")
    duration = FloatDurationField(read_only=True)

    class Meta:
        model = models.SchedulingUnitBlueprint
        fields = '__all__'
        extra_fields = ['task_blueprints', 'duration', 'start_time', 'stop_time', 'status', 'observed_end_time', 'output_pinned']
        expandable_fields = {
            'requirements_template': 'lofar.sas.tmss.tmss.tmssapp.serializers.SchedulingUnitTemplateSerializer',
            'draft': 'lofar.sas.tmss.tmss.tmssapp.serializers.SchedulingUnitDraftSerializer',
            'task_blueprints': ('lofar.sas.tmss.tmss.tmssapp.serializers.TaskBlueprintSerializer', {'many': True})
        }


class SchedulingUnitBlueprintCopyToSchedulingUnitDraftSerializer(SchedulingUnitBlueprintSerializer):
    class Meta(SchedulingUnitDraftSerializer.Meta):
       fields = ['copy_reason']
       extra_fields =['scheduling_set_id']
       read_only_fields = ['scheduling_unit_blueprints','task_drafts']


class TaskDraftSerializer(DynamicRelationalHyperlinkedModelSerializer):

    duration = FloatDurationField(read_only=True)
    relative_start_time = FloatDurationField(read_only=True)
    relative_stop_time = FloatDurationField(read_only=True)
    specifications_doc = JSONEditorField(schema_source='specifications_template.schema')

    class Meta:
        model = models.TaskDraft
        fields = '__all__'
        extra_fields = ['task_blueprints', 'produced_by', 'consumed_by', 'first_scheduling_relation', 'second_scheduling_relation', 'duration', 'relative_start_time', 'relative_stop_time']
        expandable_fields = {
            'task_blueprints': ('lofar.sas.tmss.tmss.tmssapp.serializers.TaskBlueprintSerializer', {'many': True}),
            'scheduling_unit_draft': 'lofar.sas.tmss.tmss.tmssapp.serializers.SchedulingUnitDraftSerializer',
            'specifications_template': 'lofar.sas.tmss.tmss.tmssapp.serializers.TaskTemplateSerializer'
        }


class TaskBlueprintSerializer(DynamicRelationalHyperlinkedModelSerializer):

    duration = FloatDurationField(read_only=True)
    relative_start_time = FloatDurationField(read_only=True)
    relative_stop_time = FloatDurationField(read_only=True)
    specifications_doc = JSONEditorField(schema_source='specifications_template.schema')

    class Meta:
        model = models.TaskBlueprint
        fields = '__all__'
        extra_fields = ['subtasks', 'produced_by', 'consumed_by', 'first_scheduling_relation', 'second_scheduling_relation', 'duration',
                        'start_time', 'stop_time', 'relative_start_time', 'relative_stop_time', 'status']
        expandable_fields = {
            'draft': 'lofar.sas.tmss.tmss.tmssapp.serializers.TaskDraftSerializer',
            'scheduling_unit_blueprint': 'lofar.sas.tmss.tmss.tmssapp.serializers.SchedulingUnitBlueprintSerializer',
            'specifications_template': 'lofar.sas.tmss.tmss.tmssapp.serializers.TaskTemplateSerializer',
            'subtasks': ('lofar.sas.tmss.tmss.tmssapp.serializers.SutaskSerializer', {'many': True})
        }


class TaskRelationDraftSerializer(DynamicRelationalHyperlinkedModelSerializer):
    selection_doc = JSONEditorField(schema_source='selection_template.schema')

    class Meta:
        model = models.TaskRelationDraft
        fields = '__all__'
        extra_fields = ['related_task_relation_blueprint']


class TaskRelationBlueprintSerializer(DynamicRelationalHyperlinkedModelSerializer):
    selection_doc = JSONEditorField(schema_source='selection_template.schema')

    class Meta:
        model = models.TaskRelationBlueprint
        fields = '__all__'


class TaskSchedulingRelationDraftSerializer(DynamicRelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.TaskSchedulingRelationDraft
        fields = '__all__'


class TaskSchedulingRelationBlueprintSerializer(DynamicRelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.TaskSchedulingRelationBlueprint
        fields = '__all__'


class TaskTypeSerializer(DynamicRelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.TaskType
        fields = '__all__'


class ReservationStrategyTemplateSerializer(DynamicRelationalHyperlinkedModelSerializer):
    template = JSONEditorField(schema_source="reservation_template.schema")

    class Meta:
        model = models.ReservationStrategyTemplate
        fields = '__all__'


class ReservationTemplateSerializer(AbstractTemplateSerializer):
    class Meta:
        model = models.ReservationTemplate
        fields = '__all__'


class DefaultReservationTemplateSerializer(DynamicRelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.DefaultReservationTemplate
        fields = '__all__'


class ReservationSerializer(DynamicRelationalHyperlinkedModelSerializer):
    specifications_doc = JSONEditorField(schema_source='specifications_template.schema')

    class Meta:
        model = models.Reservation
        fields = '__all__'
        extra_fields = ['duration']


class TaskBlueprintExtendedSerializer(TaskBlueprintSerializer):
    """
    Serializes not only the task blueprints, but also a subset of the referenced objects therein so that they are
    expanded into the json response for a single API call (for convenience/optimization).

    """
    subtasks = SubtaskSerializer(many=True)  # we set many=True because this field represents a sth-to-many relationship, i.e. we serialize a list of objects here
    specifications_template = TaskTemplateSerializer()


class TaskDraftExtendedSerializer(TaskDraftSerializer):
    """
    Serializes not only the task drafts, but also a subset of the referenced objects therein so that they are
    expanded into the json response for a single API call (for convenience/optimization).
    """
    task_blueprints = TaskBlueprintExtendedSerializer(many=True)
    specifications_template = TaskTemplateSerializer()


class SchedulingUnitDraftExtendedSerializer(SchedulingUnitDraftSerializer):
    """
    Serializes not only the scheduling unit drafts, but also a subset of the referenced objects therein so that they are
    expanded into the json response for a single API call (for convenience/optimization).
    """
    task_drafts = TaskDraftExtendedSerializer(many=True)


class SchedulingUnitBlueprintExtendedSerializer(SchedulingUnitBlueprintSerializer):
    """
    Serializes not only the scheduling unit blueprints, but also a subset of the referenced objects therein so that they
    are expanded into the json response for a single API call (for convenience/optimization).
    """
    task_blueprints = TaskBlueprintExtendedSerializer(many=True)
