from lofar.sas.tmss.tmss.exceptions import *
from lofar.sas.tmss.tmss.tmssapp import models
from lofar.sas.tmss.tmss.tmssapp.subtasks import unschedule_subtasks_in_task_blueprint
from lofar.sas.tmss.tmss.tmssapp.models.specification import TaskBlueprint, SchedulingUnitBlueprint
from lofar.sas.tmss.tmss.tmssapp.subtasks import create_and_schedule_subtasks_from_task_blueprint, create_subtasks_from_task_blueprint, schedule_independent_subtasks_in_task_blueprint, update_subtasks_start_times_for_scheduling_unit
from lofar.common.datetimeutils import round_to_minute_precision
from functools import cmp_to_key
import os
from copy import deepcopy
from lofar.common.json_utils import add_defaults_to_json_object_for_schema
import logging
from datetime import datetime, timedelta
from django.db.utils import IntegrityError
from django.db import transaction

logger = logging.getLogger(__name__)

def create_scheduling_unit_blueprint_from_scheduling_unit_draft(scheduling_unit_draft: models.SchedulingUnitDraft) -> models.SchedulingUnitBlueprint:
    """
    Create a SchedulingUnitBlueprint from the SchedulingUnitDraft
    :raises Exception if instantiate fails.
    """
    logger.debug("create_scheduling_unit_blueprint_from_scheduling_unit_draft(scheduling_unit_draft.id=%s name='%s')", scheduling_unit_draft.pk, scheduling_unit_draft.name)

    scheduling_unit_blueprint = SchedulingUnitBlueprint.objects.create(
        name=scheduling_unit_draft.name,
        description=scheduling_unit_draft.description,
        requirements_doc=scheduling_unit_draft.requirements_doc,
        do_cancel=False,
        draft=scheduling_unit_draft,
        requirements_template=scheduling_unit_draft.requirements_template)

    logger.info("create_scheduling_unit_blueprint_from_scheduling_unit_draft(scheduling_unit_draft.id=%s name='%s') created scheduling_unit_blueprint id=%s name='%s'",
                scheduling_unit_draft.pk, scheduling_unit_draft.name, scheduling_unit_blueprint.pk, scheduling_unit_blueprint.name)
    return scheduling_unit_blueprint


def copy_scheduling_unit_draft(scheduling_unit_draft: models.SchedulingUnitDraft, scheduling_set_dst: models.SchedulingSet, copy_reason: str) -> models.SchedulingUnitDraft:
    """
    Copy a SchedulingUnitDraft
    :raises Exception if instantiate fails.
    """
    task_drafts = list(scheduling_unit_draft.task_drafts.all())
    scheduling_unit_draft_copy = deepcopy(scheduling_unit_draft)
    scheduling_unit_draft_copy.id = None
    scheduling_unit_draft_copy.copies=scheduling_unit_draft
    if copy_reason is not None:
        scheduling_unit_draft_copy.copy_reason = models.CopyReason.objects.get(value=copy_reason)

    scheduling_unit_draft_copy.name="%s (Copy)" % (scheduling_unit_draft.name,)
    scheduling_unit_draft_copy.description="%s (Copy from %s)" % (scheduling_unit_draft.description or "<no description>",scheduling_unit_draft.name,)

    scheduling_unit_draft_copy.scheduling_set=scheduling_set_dst

    task_drafts_copy = []
    scheduling_unit_draft_copy.save()
    for td in task_drafts:
        task_drafts_copy.append(copy_task_draft(td))
    scheduling_unit_draft_copy.task_drafts.set(task_drafts_copy)
    scheduling_unit_draft_copy.save()

    logger.info("copy_scheduling_unit_draft(scheduling_unit_draft.id=%s) created copy_scheduling_unit_draft id=%s", scheduling_unit_draft.pk, scheduling_unit_draft_copy.pk)
    return scheduling_unit_draft_copy


def create_scheduling_unit_draft_from_scheduling_unit_blueprint(scheduling_unit_blueprint: models.SchedulingUnitBlueprint, copy_reason: str) -> models.SchedulingUnitDraft:
    """
    Create a SchedulingUnitDraft from the SchedulingUnitBlueprint
     :raises Exception if instantiate fails.
    """
    logger.debug("create_scheduling_unit_draft_from_scheduling_unit_blueprint(scheduling_unit_blueprint.id=%s)", scheduling_unit_blueprint.pk)

    original_scheduling_unit_draft = scheduling_unit_blueprint.draft

    # TODO: fix the copy reason
    scheduling_unit_draft =  SchedulingUnitDraft.objects.create(
        name="%s (Copied from SchedulingUnitBlueprint)" % (scheduling_unit_blueprint.name,),
        description="%s (Copied from %s SchedulingUnitBlueprint)" % (scheduling_unit_blueprint.description or "<no description>",scheduling_unit_blueprint.name,),
        requirements_doc=scheduling_unit_blueprint.requirements_doc,
        copy_reason=copy_reason,
        generator_instance_doc=original_scheduling_unit_draft.generator_instance_doc,
        scheduling_set=original_scheduling_unit_draft.scheduling_set,
        requirements_template=scheduling_unit_blueprint.requirements_template)

    task_drafts_copy = []
    task_blueprints = list(scheduling_unit_blueprint.task_blueprints.all())

    for tb in task_blueprints:
        task_drafts_copy.append(copy_task_blueprint_to_task_draft(tb))
    scheduling_unit_draft.task_drafts.set(task_drafts_copy)
    scheduling_unit_draft.save()

    logger.info("create_scheduling_unit_draft_from_scheduling_unit_blueprint(scheduling_unit_blueprint.id=%s) created scheduling_unit_draft id=%s", scheduling_unit_blueprint.pk, scheduling_unit_draft.pk)
    return scheduling_unit_draft


def copy_task_draft(task_draft: models.TaskDraft, copy_reason: str) -> models.TaskDraft:

    task_template_name = task_draft.specifications_template
    task_template = models.TaskTemplate.objects.get(name=task_template_name)

    task_draft_copy = models.TaskDraft.objects.create(
                                name=task_draft.name,
                                description=task_draft.description,
                                tags=task_draft.tags,
                                specifications_doc=task_draft.specifications_doc,
                                copy_reason=copy_reason,
                                copies=task_draft,
                                scheduling_unit_draft=task_draft.scheduling_unit_draft,
                                specifications_template=task_template)
    logger.info("task draft with id %s created successfully" % task_draft.id)
    logger.info("copy_task_draft(task_draft.id=%s) created copy_task_draft id=%s", task_draft.pk, task_draft_copy.pk)
    return task_draft_copy


def copy_task_blueprint_to_task_draft(task_blueprint:models.TaskBlueprint ) -> models.TaskDraft:
    """
    Create a copy of a task_blueprint to a task_draft
    :raises Exception if instantiate fails.
    """
    logger.debug("Create Task Draft  from Task Blueprint (id=%s)", task_blueprint.pk)

    original_task_draft = task_blueprint.draft
    task_template_name = original_task_draft.specifications_template
    task_template = models.TaskTemplate.objects.get(name=task_template_name)

    task_draft_copy = models.TaskDraft.objects.create(
            name="%s (Copied from Task Blueprint)" % (task_blueprint.name,),
            description="%s (Copied from %s Task Blueprint)" % (task_blueprint.description or "<no description>",task_blueprint.name,),
            tags=original_task_draft.tags,
            specifications_doc=original_task_draft.specifications_doc,
            copy_reason=original_task_draft.copy_reason,
            copies=original_task_draft,
            scheduling_unit_draft=original_task_draft.scheduling_unit_draft,
            specifications_template=task_template)
    logger.info("copy_task_blueprint_to_task_draft(task_blueprint.id=%s) created task_draft id=%s", task_blueprint.pk, task_draft_copy.pk)
    return task_draft_copy


def create_task_drafts_from_scheduling_unit_draft(scheduling_unit_draft: models.SchedulingUnitDraft) -> models.SchedulingUnitDraft:
    """
    Create all defined task_drafts in the scheduling_unit_draft's requirements_doc, connect them, and return the updated scheduling_unit_draft.
    """
    logger.debug("create_task_drafts_from_scheduling_unit_draft(scheduling_unit_draft.id=%s, name='%s') ...", scheduling_unit_draft.pk, scheduling_unit_draft.name)

    if len(scheduling_unit_draft.requirements_doc.get("tasks", {})) == 0:
        raise BlueprintCreationException("create_task_drafts_from_scheduling_unit_draft: scheduling_unit_draft.id=%s has no tasks defined in its requirements_doc" % (scheduling_unit_draft.pk,))

    for task_name, task_definition in scheduling_unit_draft.requirements_doc["tasks"].items():
        task_template_name = task_definition["specifications_template"]
        task_template = models.TaskTemplate.objects.get(name=task_template_name)

        task_specifications_doc = task_definition["specifications_doc"]
        task_specifications_doc = add_defaults_to_json_object_for_schema(task_specifications_doc, task_template.schema)

        try:
            logger.debug("creating task draft... task_name='%s', task_template_name='%s'", task_template_name, task_template_name)

            with transaction.atomic():
                task_draft = models.TaskDraft.objects.create(name=task_name,
                                                             description=task_definition.get("description",""),
                                                             tags=task_definition.get("tags",[]),
                                                             specifications_doc=task_specifications_doc,
                                                             copy_reason=models.CopyReason.objects.get(value='template'),
                                                             copies=None,
                                                             scheduling_unit_draft=scheduling_unit_draft,
                                                             specifications_template=task_template)

                logger.info("created task draft id=%s task_name='%s', task_template_name='%s'", task_draft.pk, task_name, task_template_name)
        except IntegrityError as e:
            if 'TaskDraft_unique_name_in_scheduling_unit' in str(e):
                logger.info("task draft task_name='%s', task_template_name='%s' already exists in scheduling_unit id=%s name='%s'",
                            task_name, task_template_name, scheduling_unit_draft.id, scheduling_unit_draft.name)
            else:
                raise

    # Now create task relations
    for task_relation_definition in scheduling_unit_draft.requirements_doc["task_relations"]:
        producer_task_draft = scheduling_unit_draft.task_drafts.get(name=task_relation_definition["producer"])
        consumer_task_draft = scheduling_unit_draft.task_drafts.get(name=task_relation_definition["consumer"])
        dataformat = models.Dataformat.objects.get(value=task_relation_definition["dataformat"])
        input_role = models.TaskConnectorType.objects.get(task_template=consumer_task_draft.specifications_template, role=task_relation_definition["input"]["role"], datatype=task_relation_definition["input"]["datatype"], iotype=models.IOType.objects.get(value=models.IOType.Choices.INPUT.value))
        output_role = models.TaskConnectorType.objects.get(task_template=producer_task_draft.specifications_template, role=task_relation_definition["output"]["role"], datatype=task_relation_definition["output"]["datatype"], iotype=models.IOType.objects.get(value=models.IOType.Choices.OUTPUT.value))
        selection_template = models.TaskRelationSelectionTemplate.objects.get(name=task_relation_definition["selection_template"])

        try:
            with transaction.atomic():
                task_relation = models.TaskRelationDraft.objects.create(tags=task_relation_definition.get("tags",[]),
                                                                        selection_doc=task_relation_definition["selection_doc"],
                                                                        dataformat=dataformat,
                                                                        producer=producer_task_draft,
                                                                        consumer=consumer_task_draft,
                                                                        input_role=input_role,
                                                                        output_role=output_role,
                                                                        selection_template=selection_template)
                logger.info("created task_relation id=%s between task draft id=%s name='%s' and id=%s name='%s",
                            task_relation.pk, producer_task_draft.id, producer_task_draft.name, consumer_task_draft.id, consumer_task_draft.name)
        except IntegrityError as e:
            if 'TaskRelationDraft_unique_relation' in str(e):
                logger.info("task_relation between task draft id=%s name='%s' and id=%s name='%s already exists",
                             producer_task_draft.id, producer_task_draft.name, consumer_task_draft.id, consumer_task_draft.name)
            else:
                raise


    # task_scheduling_relation
    for task_scheduling_relation_definition in scheduling_unit_draft.requirements_doc["task_scheduling_relations"]:
        placement = models.SchedulingRelationPlacement.objects.get(value=task_scheduling_relation_definition["placement"])
        time_offset = task_scheduling_relation_definition["time_offset"]
        first_task_draft = scheduling_unit_draft.task_drafts.get(name=task_scheduling_relation_definition["first"])
        second_task_draft = scheduling_unit_draft.task_drafts.get(name=task_scheduling_relation_definition["second"])

        try:
            with transaction.atomic():
                task_scheduling_relation = models.TaskSchedulingRelationDraft.objects.create(placement=placement,
                                                                                             time_offset=time_offset,
                                                                                             first=first_task_draft,
                                                                                             second=second_task_draft)
                logger.info("created task_scheduling_relation id=%s between task draft id=%s name='%s' and id=%s name='%s",
                            task_scheduling_relation.pk, first_task_draft.id, first_task_draft.name, second_task_draft.id, second_task_draft.name)
        except IntegrityError as e:
            if 'TaskSchedulingRelationDraft_unique_relation' in str(e):
                logger.info("task_scheduling_relation between task draft id=%s name='%s' and id=%s name='%s already exists",
                            first_task_draft.id, first_task_draft.name, second_task_draft.id, second_task_draft.name)
            else:
                raise

    scheduling_unit_draft.refresh_from_db()
    return scheduling_unit_draft


def create_task_blueprint_from_task_draft(task_draft: models.TaskDraft) -> models.TaskBlueprint:
    """
    Create a task_blueprint from the task_draft
    :raises Exception if instantiate fails.
    """
    logger.debug("creating task_blueprint from task_draft id=%s", task_draft.pk)

    # get or create a scheduling_unit_blueprint from the scheduling_unit_draft
    scheduling_unit_blueprint = task_draft.scheduling_unit_draft.scheduling_unit_blueprints.last()
    if scheduling_unit_blueprint is None:
        scheduling_unit_blueprint = create_scheduling_unit_blueprint_from_scheduling_unit_draft(task_draft.scheduling_unit_draft)

    try:
        with transaction.atomic():
            task_blueprint = TaskBlueprint.objects.create(
            description=task_draft.description,
            name=task_draft.name,
            do_cancel=False,
            draft=task_draft,
            scheduling_unit_blueprint=scheduling_unit_blueprint,
            specifications_doc=task_draft.specifications_doc,
            specifications_template=task_draft.specifications_template,
            output_pinned=task_draft.output_pinned)

            logger.info("created task_blueprint id=%s from task_draft id=%s", task_blueprint.pk, task_draft.pk)
    except IntegrityError as e:
        if 'TaskBlueprint_unique_name_in_scheduling_unit' in str(e) or 'TaskBlueprint_unique_from_task_draft_in_scheduling_unit' in str(e):
            logger.info("task_blueprint from task_draft id=%s already exists in scheduling_unit_blueprint id=%s name='%s'",
                        task_draft.pk, scheduling_unit_blueprint.id, scheduling_unit_blueprint.name)
            # todo: return the existing blueprint or meaningful exception?
            #  (Currently we get: local variable 'task_blueprint' referenced before assignment)
        else:
            raise

    # now that we have a task_blueprint, its time to refresh the task_draft so we get the non-cached fields
    task_draft.refresh_from_db()

    # loop over consumers/producers, and 'copy' the TaskRelationBlueprint from the TaskRelationDraft
    # this is only possible if both 'ends' of the task_relation are converted to a TaskBlueprint
    # so, when converting two TaskDrafts (for example an observation and a pipeline), then for the conversion
    # of the first TaskDraft->TaskBlueprint no relation is setup,
    # and for the second TaskDraft->TaskBlueprint conversion we have both endpoints available, so we can connect them.
    task_draft_relations = list(task_draft.consumed_by.all()) + list(task_draft.produced_by.all())
    for task_relation_draft in task_draft_relations:
        for producing_task_blueprint in task_relation_draft.producer.task_blueprints.all():
            for consuming_task_blueprint in task_relation_draft.consumer.task_blueprints.all():
                try:
                    with transaction.atomic():
                        task_relation_blueprint = models.TaskRelationBlueprint.objects.create(draft=task_relation_draft,
                                                                                              input_role=task_relation_draft.input_role,
                                                                                              output_role=task_relation_draft.output_role,
                                                                                              producer=producing_task_blueprint,
                                                                                              consumer=consuming_task_blueprint,
                                                                                              selection_doc=task_relation_draft.selection_doc,
                                                                                              selection_template=task_relation_draft.selection_template,
                                                                                              dataformat=task_relation_draft.dataformat)
                        logger.info("created task_relation_blueprint id=%s which connects task_blueprints producer_id=%s and consumer_id=%s",
                                    task_relation_blueprint.pk, producing_task_blueprint.pk, consuming_task_blueprint.pk)
                except IntegrityError as e:
                    if 'TaskRelationBlueprint_unique_relation' in str(e):
                        logger.info("task_relation_blueprint with producer_id=%s and consumer_id=%s already exists",
                                    producing_task_blueprint.pk, consuming_task_blueprint.pk)
                    else:
                        raise


    # Do the same 'trick' for Task Scheduling Relation Draft to Blueprint
    task_draft_scheduling_relations = list(task_draft.first_scheduling_relation.all()) + list(task_draft.second_scheduling_relation.all())
    for task_scheduling_relation_draft in task_draft_scheduling_relations:
        for first_task_blueprint in task_scheduling_relation_draft.first.task_blueprints.all():
            for second_task_blueprint in task_scheduling_relation_draft.second.task_blueprints.all():
                try:
                    with transaction.atomic():
                        task_scheduling_relation_blueprint = models.TaskSchedulingRelationBlueprint.objects.create(first=first_task_blueprint,
                                                                                                                   second=second_task_blueprint,
                                                                                                                   time_offset=task_scheduling_relation_draft.time_offset,
                                                                                                                   placement=task_scheduling_relation_draft.placement)
                        logger.info("created task_scheduling_relation_blueprint id=%s which connects task_blueprints first_id=%s and second_id=%s, placement=%s time_offset=%s[sec]",
                                    task_scheduling_relation_blueprint.pk, first_task_blueprint.pk, second_task_blueprint.pk, task_scheduling_relation_draft.placement, task_scheduling_relation_draft.time_offset)
                except IntegrityError as e:
                    if 'TaskSchedulingRelationBlueprint_unique_relation' in str(e):
                        logger.info("task_scheduling_relation_blueprint with producer_id=%s and consumer_id=%s already exists",
                                    producing_task_blueprint.pk, consuming_task_blueprint.pk)
                    else:
                        raise

    return task_blueprint


def create_task_blueprints_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft: models.SchedulingUnitDraft) -> models.SchedulingUnitBlueprint:
    '''Convenience method: Create the scheduling_unit_blueprint, then create its child task_blueprint(s), then create the task_blueprint's subtasks'''

    # make sure we call create task_drafts first, so the task_blueprints can be created from the scheduling_unit_blueprint in the next step
    # already known task_drafts and/or relations are skipped automagically
    create_task_drafts_from_scheduling_unit_draft(scheduling_unit_draft)

    scheduling_unit_blueprint = create_scheduling_unit_blueprint_from_scheduling_unit_draft(scheduling_unit_draft)
    return create_task_blueprints_and_subtasks_from_scheduling_unit_blueprint(scheduling_unit_blueprint)


def create_task_blueprints_and_subtasks_and_schedule_subtasks_from_scheduling_unit_draft(scheduling_unit_draft: models.SchedulingUnitDraft) -> models.SchedulingUnitBlueprint:
    '''Convenience method: Create the scheduling_unit_blueprint, then create its child task_blueprint(s), then create the task_blueprint's subtasks, and schedule the ones that are not dependend on predecessors'''
    scheduling_unit_blueprint = create_scheduling_unit_blueprint_from_scheduling_unit_draft(scheduling_unit_draft)
    return create_task_blueprints_and_subtasks_and_schedule_subtasks_from_scheduling_unit_blueprint(scheduling_unit_blueprint)


def create_task_blueprint_and_subtasks_from_task_draft(task_draft: models.TaskDraft) -> models.TaskBlueprint:
    '''Convenience method: Create the task_blueprint, then create the task_blueprint's subtasks, and schedule the ones that are not dependend on predecessors'''
    task_blueprint =  create_task_blueprint_from_task_draft(task_draft)
    create_subtasks_from_task_blueprint(task_blueprint)
    task_blueprint.refresh_from_db()
    return task_blueprint


def create_task_blueprint_and_subtasks_and_schedule_subtasks_from_task_draft(task_draft: models.TaskDraft) -> models.TaskBlueprint:
    '''Convenience method: Create the task_blueprint, then create the task_blueprint's subtasks, and schedule the ones that are not dependend on predecessors'''
    task_blueprint =  create_task_blueprint_from_task_draft(task_draft)
    create_and_schedule_subtasks_from_task_blueprint(task_blueprint)
    task_blueprint.refresh_from_db()
    return task_blueprint


def create_task_blueprints_from_scheduling_unit_blueprint(scheduling_unit_blueprint: models.SchedulingUnitBlueprint) -> models.SchedulingUnitBlueprint:
    '''Convenience method: Create the scheduling_unit_blueprint's task_blueprint(s)'''

    # make sure we call create task_drafts first, so the task_blueprints can be created from the scheduling_unit_blueprint in the next step
    # already known task_drafts and/or relations are skipped automagically
    create_task_drafts_from_scheduling_unit_draft(scheduling_unit_blueprint.draft)

    task_drafts = list(scheduling_unit_blueprint.draft.task_drafts.all())

    # sort them in 'data-flow'-order,
    # because successors can depend on predecessors, so the first taskdraft's need to be blueprinted first.
    task_drafts.sort(key=cmp_to_key(lambda taskdraft_a, taskdraft_b: -1 if taskdraft_a in taskdraft_b.predecessors else 1 if taskdraft_b in taskdraft_a.predecessors else 0))

    # convert task_draft(s) to task_blueprint(s)
    for task_draft in task_drafts:
        create_task_blueprint_from_task_draft(task_draft)

    # refresh so all related fields are updated.
    scheduling_unit_blueprint.refresh_from_db()
    return scheduling_unit_blueprint


def create_task_blueprints_and_subtasks_from_scheduling_unit_blueprint(scheduling_unit_blueprint: models.SchedulingUnitBlueprint) -> models.SchedulingUnitBlueprint:
    '''Convenience method: Create the scheduling_unit_blueprint's task_blueprint(s), then create each task_blueprint's subtasks'''
    scheduling_unit_blueprint = create_task_blueprints_from_scheduling_unit_blueprint(scheduling_unit_blueprint)

    for task_blueprint in scheduling_unit_blueprint.task_blueprints.all():
        create_subtasks_from_task_blueprint(task_blueprint)

    # assign reasonable default start/stop times so the subtasks/tasks/sched_unit can be displayed in a timeline view
    # these default start/stop times can of course be overridden by the operator and/or dynamic scheduling.
    update_subtasks_start_times_for_scheduling_unit(scheduling_unit_blueprint, round_to_minute_precision(datetime.utcnow()+timedelta(hours=1)))

    # refresh so all related fields are updated.
    scheduling_unit_blueprint.refresh_from_db()

    return scheduling_unit_blueprint


def create_task_blueprints_and_subtasks_and_schedule_subtasks_from_scheduling_unit_blueprint(scheduling_unit_blueprint: models.SchedulingUnitBlueprint) -> models.SchedulingUnitBlueprint:
    '''Convenience method: Create the scheduling_unit_blueprint's task_blueprint(s), then create the task_blueprint's subtasks, and schedule the ones that are not dependend on predecessors'''
    scheduling_unit_blueprint = create_task_blueprints_and_subtasks_from_scheduling_unit_blueprint(scheduling_unit_blueprint)
    scheduling_unit_blueprint = schedule_independent_subtasks_in_scheduling_unit_blueprint(scheduling_unit_blueprint)
    return scheduling_unit_blueprint


def schedule_independent_subtasks_in_scheduling_unit_blueprint(scheduling_unit_blueprint: SchedulingUnitBlueprint, start_time: datetime=None) -> models.SchedulingUnitBlueprint:
    '''Convenience method: Schedule the subtasks in the scheduling_unit_blueprint that are not dependend on predecessors'''
    task_blueprints = list(scheduling_unit_blueprint.task_blueprints.all())

    for task_blueprint in task_blueprints:
        schedule_independent_subtasks_in_task_blueprint(task_blueprint, start_time=start_time+task_blueprint.relative_start_time)

    scheduling_unit_blueprint.refresh_from_db()
    return scheduling_unit_blueprint


def unschedule_subtasks_in_scheduling_unit_blueprint(scheduling_unit_blueprint: SchedulingUnitBlueprint) -> models.SchedulingUnitBlueprint:
    '''Convenience method: Unschedule all scheduled subtasks in the scheduling_unit_blueprint'''
    task_blueprints = list(scheduling_unit_blueprint.task_blueprints.all())

    for task_blueprint in task_blueprints:
        unschedule_subtasks_in_task_blueprint(task_blueprint)

    scheduling_unit_blueprint.refresh_from_db()
    return scheduling_unit_blueprint

