#!/usr/bin/env python3

# Copyright (C) 2012-2015  ASTRON (Netherlands Institute for Radio Astronomy)
# P.O. Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.


import os
import unittest
import requests

import logging
logger = logging.getLogger(__name__)
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=logging.INFO)

from lofar.common.test_utils import exit_with_skipped_code_if_skip_integration_tests
exit_with_skipped_code_if_skip_integration_tests()

from lofar.common.json_utils import get_default_json_object_for_schema, add_defaults_to_json_object_for_schema


# Do Mandatory setup step:
# use setup/teardown magic for tmss test database, ldap server and django server
# (ignore pycharm unused import statement, python unittests does use at RunTime the tmss_test_environment_unittest_setup module)
from lofar.sas.tmss.test.tmss_test_environment_unittest_setup import *
tmss_test_env.populate_schemas()

from lofar.sas.tmss.test.tmss_test_data_django_models import *

# import and setup rest test data creator
from lofar.sas.tmss.test.tmss_test_data_rest import TMSSRESTTestDataCreator
rest_data_creator = TMSSRESTTestDataCreator(BASE_URL, AUTH)

from lofar.sas.tmss.tmss.tmssapp import models


from lofar.sas.tmss.tmss.tmssapp.reservations import get_active_station_reservations_in_timewindow


class TestStationReservations(unittest.TestCase):
    """
    Tests for the active station reservations
    """

    def setUp(self) -> None:
         # wipe all reservations in between tests, so the tests don't influence each other
        for reservation in models.Reservation.objects.all():
            reservation.delete()

    @staticmethod
    def create_station_reservation(additional_name, lst_stations, start_time, stop_time=None):
        """
        Create a station reservation with given list of stations, start_time and stop_time
        """
        reservation_template = models.ReservationTemplate.objects.get(name="resource reservation")
        reservation_template_spec = get_default_json_object_for_schema(reservation_template.schema)
        reservation_template_spec["resources"] = {"stations": lst_stations }
        res = models.Reservation.objects.create(name="Station Reservation %s" % additional_name,
                                   description="Station reservation for testing",
                                   specifications_template=reservation_template,
                                   specifications_doc=reservation_template_spec,
                                   start_time=start_time,
                                   stop_time=stop_time)

    def test_no_stations_reservation(self):
        """
        Check that creating 'default' reservation with no additional station reservation added, we still can
        call 'get_active_station_reservations_in_timewindow' and it will return an empty list
        """
        reservation_template = models.ReservationTemplate.objects.get(name="resource reservation")
        reservation_template_spec = get_default_json_object_for_schema(reservation_template.schema)
        res = models.Reservation.objects.create(name="AnyReservation",
                                   description="Reservation of something else",
                                   specifications_template=reservation_template,
                                   specifications_doc=reservation_template_spec,
                                   start_time=datetime.now(),
                                   stop_time=None)
        self.assertCountEqual([],
                              get_active_station_reservations_in_timewindow(datetime.now(), datetime.now()+timedelta(weeks=53)))

    def test_active_station_reservation(self):
        """
        Test station reservation when 2 stations are reserved for 24hr with both same start and stop time
        Check 'get_active_station_reservations_in_timewindow' with different time ranges
        """
        reservation_start_time = datetime(2020, 1, 1, 0, 0, 0)
        reservation_stop_time = datetime(2020, 1, 2, 0, 0, 0)
        reservation_stations = ["CS001", "CS002"]
        self.create_station_reservation("two_stations", reservation_stations, reservation_start_time, reservation_stop_time)
        self.assertCountEqual([],
                              get_active_station_reservations_in_timewindow(datetime.now(), datetime.now()))
        self.assertCountEqual(reservation_stations,
                              get_active_station_reservations_in_timewindow(reservation_start_time,
                                                                            reservation_stop_time))
        self.assertCountEqual([],
                              get_active_station_reservations_in_timewindow(reservation_start_time,
                                                                            reservation_start_time))
        self.assertCountEqual(reservation_stations,
                              get_active_station_reservations_in_timewindow(reservation_start_time,
                                                                            reservation_start_time+timedelta(seconds=1)))
        self.assertCountEqual([],
                              get_active_station_reservations_in_timewindow(reservation_stop_time,
                                                                            reservation_stop_time))
        self.assertCountEqual(reservation_stations,
                              get_active_station_reservations_in_timewindow(reservation_stop_time-timedelta(seconds=1),
                                                                            reservation_stop_time))
        self.assertCountEqual(reservation_stations,
                              get_active_station_reservations_in_timewindow(reservation_start_time-timedelta(weeks=53),
                                                                            reservation_stop_time+timedelta(weeks=53)))
        self.assertCountEqual(reservation_stations,
                              get_active_station_reservations_in_timewindow(reservation_start_time,
                                                                            datetime.now()+timedelta(weeks=53)))
        self.assertCountEqual(reservation_stations,
                              get_active_station_reservations_in_timewindow(datetime(2020, 1, 1, 12, 0, 0),
                                                                            datetime(2020, 1, 2, 12, 0, 0)))

    def test_active_station_reservation_with_same_station_overlap(self):
        """
        Test station reservation when 2 stations are reserved for 24hr with both same start and stop time
        Same stations are reserved separately as well, which should result NOT result in having the station twice in
        the resulting active station list
        Check 'get_active_station_reservations_in_timewindow' with different time ranges
        """
        reservation_start_time = datetime(2020, 1, 1, 0, 0, 0)
        reservation_stop_time = datetime(2020, 1, 2, 0, 0, 0)
        reservation_stations = ["CS001", "CS002"]
        self.create_station_reservation("two_stations", reservation_stations, reservation_start_time, reservation_stop_time)
        self.create_station_reservation("cs1", ["CS001"], reservation_start_time, reservation_stop_time)
        self.create_station_reservation("cs2", ["CS002"], reservation_start_time, reservation_stop_time)

        # same lower_bound as upper_bound, empty list
        self.assertCountEqual([],
                              get_active_station_reservations_in_timewindow(datetime.now(), datetime.now()))
        self.assertCountEqual(reservation_stations,
                              get_active_station_reservations_in_timewindow(reservation_start_time,
                                                                            reservation_stop_time))
        self.assertCountEqual([],
                              get_active_station_reservations_in_timewindow(reservation_start_time,
                                                                            reservation_start_time))
        self.assertCountEqual(reservation_stations,
                              get_active_station_reservations_in_timewindow(reservation_start_time,
                                                                            reservation_start_time + timedelta(seconds=1)))
        self.assertCountEqual([],
                              get_active_station_reservations_in_timewindow(reservation_stop_time,
                                                                            reservation_stop_time + timedelta(seconds=1)))
        self.assertCountEqual(reservation_stations,
                              get_active_station_reservations_in_timewindow(reservation_stop_time - timedelta(seconds=1),
                                                                            reservation_stop_time))
        self.assertCountEqual(reservation_stations,
                              get_active_station_reservations_in_timewindow(reservation_start_time - timedelta(weeks=53),
                                                                            reservation_stop_time + timedelta(weeks=53)))
        self.assertCountEqual(reservation_stations,
                              get_active_station_reservations_in_timewindow(reservation_start_time,
                                                                            datetime.now() + timedelta(weeks=53)))
        self.assertCountEqual(reservation_stations,
                              get_active_station_reservations_in_timewindow(datetime(2020, 1, 1, 12, 0, 0),
                                                                            datetime(2020, 1, 2, 12, 0, 0)))

    def test_active_station_reservation_with_station_no_stop_time(self):
        """
        Test station reservation when 2 stations are reserved forever both same start and stop time
        Check 'get_active_station_reservations_in_timewindow' with different time ranges
        """
        reservation_start_time = datetime(2020, 1, 1, 0, 0, 0)
        reservation_stations = ["CS001", "CS002"]
        self.create_station_reservation("two_stations_no_end_time", reservation_stations, reservation_start_time)

        # we are now still in reservation started in 2020
        self.assertCountEqual(reservation_stations,
                              get_active_station_reservations_in_timewindow(datetime.now(), datetime.now()))
        self.assertCountEqual(reservation_stations,
                              get_active_station_reservations_in_timewindow(reservation_start_time,
                                                                            reservation_start_time + timedelta(seconds=1)))
        # before start time, always empty
        self.assertCountEqual([],
                              get_active_station_reservations_in_timewindow(reservation_start_time - timedelta(seconds=1),
                                                                            reservation_start_time))
        self.assertCountEqual(reservation_stations,
                              get_active_station_reservations_in_timewindow(reservation_start_time - timedelta(weeks=53),
                                                                            reservation_start_time + timedelta(weeks=53)))
        self.assertCountEqual(reservation_stations,
                              get_active_station_reservations_in_timewindow(reservation_start_time,
                                                                            datetime.now() + timedelta(weeks=53)))
        self.assertCountEqual(reservation_stations,
                              get_active_station_reservations_in_timewindow(datetime(2020, 1, 1, 12, 0, 0),
                                                                            datetime(2020, 1, 2, 12, 0, 0)))

    def test_active_station_reservation_every_hour_one_station(self):
        """
        Test station reservation when 1 station is reserved for 1 hour
        Check 'get_active_station_reservations_in_timewindow' with different time ranges
        """
        first_day = 1
        last_day = 7
        reservation_start_time = datetime(2020, 1, first_day, 0, 0, 0)
        reservation_stop_time = datetime(2020, 1, last_day+1, 0, 0, 0)
        for day_nbr in range(first_day, last_day+1):
            self.create_station_reservation("cs%s" % day_nbr, ["CS00%d" % day_nbr],
                                            datetime(2020, 1, day_nbr, 12, 0, 0), datetime(2020, 1, day_nbr+1, 0, 0, 0))

        self.assertCountEqual([],
                              get_active_station_reservations_in_timewindow(datetime.now(), datetime.now()))
        self.assertCountEqual(["CS001","CS002","CS003","CS004","CS005","CS006","CS007"],
                              get_active_station_reservations_in_timewindow(reservation_start_time, reservation_stop_time))
        self.assertCountEqual(["CS002","CS003","CS004","CS005","CS006","CS007"],
                              get_active_station_reservations_in_timewindow(reservation_start_time+timedelta(days=1), reservation_stop_time))
        self.assertCountEqual(["CS003","CS004","CS005","CS006","CS007"],
                              get_active_station_reservations_in_timewindow(reservation_start_time+timedelta(days=2), reservation_stop_time))
        self.assertCountEqual(["CS004","CS005","CS006","CS007"],
                              get_active_station_reservations_in_timewindow(reservation_start_time+timedelta(days=3), reservation_stop_time))
        self.assertCountEqual(["CS005","CS006","CS007"],
                              get_active_station_reservations_in_timewindow(reservation_start_time+timedelta(days=4), reservation_stop_time))
        self.assertCountEqual(["CS006","CS007"],
                              get_active_station_reservations_in_timewindow(reservation_start_time+timedelta(days=5), reservation_stop_time))
        self.assertCountEqual(["CS007"],
                              get_active_station_reservations_in_timewindow(reservation_start_time+timedelta(days=6), reservation_stop_time))

        self.assertCountEqual(["CS001","CS002","CS003","CS004","CS005","CS006"],
                              get_active_station_reservations_in_timewindow(reservation_start_time, reservation_stop_time-timedelta(days=1)))
        self.assertCountEqual(["CS001","CS002","CS003","CS004","CS005"],
                              get_active_station_reservations_in_timewindow(reservation_start_time, reservation_stop_time-timedelta(days=2)))
        self.assertCountEqual(["CS001","CS002","CS003","CS004"],
                              get_active_station_reservations_in_timewindow(reservation_start_time, reservation_stop_time-timedelta(days=3)))
        self.assertCountEqual(["CS001","CS002","CS003"],
                              get_active_station_reservations_in_timewindow(reservation_start_time, reservation_stop_time-timedelta(days=4)))
        self.assertCountEqual(["CS001","CS002"],
                              get_active_station_reservations_in_timewindow(reservation_start_time, reservation_stop_time-timedelta(days=5)))
        self.assertCountEqual(["CS001"],
                              get_active_station_reservations_in_timewindow(reservation_start_time, reservation_stop_time-timedelta(days=6)))
from lofar.sas.tmss.tmss.exceptions import SchemaValidationException
from django.core.exceptions import ValidationError


class CreationFromReservationStrategyTemplate(unittest.TestCase):
    """
    Test that reservations can be created from strategy template
    """

    def test_create_reservation_ok(self):
        """
        Check that reservations from the reservation strategy can be created with api
        """
        strategy_template = models.ReservationStrategyTemplate.objects.get(name="Regular station maintenance")

        reservation_spec = add_defaults_to_json_object_for_schema(strategy_template.template,
                                                                  strategy_template.reservation_template.schema)
        reservation = models.Reservation.objects.create(name=strategy_template.name,
                                                 description="Unittest with %s" % strategy_template.description,
                                                 project=None,
                                                 specifications_template=strategy_template.reservation_template,
                                                 specifications_doc=reservation_spec,
                                                 start_time=datetime.now(),
                                                 stop_time=None)

        # Check URL of the reservation that is created
        response = GET_and_assert_equal_expected_code(self, BASE_URL + '/reservation/%d' % reservation.pk, 200)
        self.assertEqual(response['id'], reservation.pk)  # should be different id then previous one created

        # Check that action call 'create_reservation' (no parameters) of strategy template creates a
        # new reservation (with http result code 201)
        response = GET_and_assert_equal_expected_code(self, BASE_URL + '/reservation_strategy_template/%d/create_reservation' % strategy_template.pk, 201)
        self.assertNotEqual(response['id'], reservation.pk)  # should be different id then previous one created
        self.assertLess(response['start_time'], datetime.utcnow().isoformat())   # start_time created with now so that was some micro seconds ago
        self.assertEqual(response['stop_time'], None)
        self.assertEqual(response['duration'], None)
        self.assertEqual(response['name'], "reservation")
        self.assertEqual(response['specifications_doc'], reservation_spec)

    def test_create_reservation_exception(self):
        """
        Check that reservations from the reservation strategy results in an Exception due to wrong
        station assignment
        """
        strategy_template = models.ReservationStrategyTemplate.objects.get(name="Regular station maintenance")
        strategy_template.template['resources']['stations'] = ['CS999']
        # using ValidationError seem not  to work?
        with self.assertRaises(Exception) as context:
            strategy_template.save()
        self.assertIn('is not one of', str(context.exception))
        self.assertIn('Failed validating', str(context.exception))


class ReservationTest(unittest.TestCase):
    """
    Check the Reservation model
    TODO: more testcases to be added
    """

    def test_create_reservation_validation_error(self):
        """
        Check that creating reservation with results in SchemaValidationException due to wrong station assignment
        """
        reservation_template = models.ReservationTemplate.objects.get(pk=1)
        reservation_spec = get_default_json_object_for_schema(reservation_template.schema)
        reservation_spec['resources']['stations'] = ['CS999']
        with self.assertRaises(SchemaValidationException) as context:
            models.Reservation.objects.create(name="Test Reservation",
                                              description="Unittest",
                                              project=None,
                                              specifications_template=reservation_template,
                                              specifications_doc=reservation_spec,
                                              start_time=datetime.now(),
                                              stop_time=None)
        self.assertIn('is not one of', str(context.exception))

